# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionalities for interpolating a regular grid at points (raster to point cloud)."""

from __future__ import annotations

from typing import Any, Callable, Literal, overload

import numpy as np
import rasterio as rio
from scipy.interpolate import RectBivariateSpline, RegularGridInterpolator
from scipy.ndimage import binary_dilation, distance_transform_edt, map_coordinates

from geoutils import profiler
from geoutils._typing import NDArrayNum, Number
from geoutils.raster.georeferencing import _coords, _outside_image, _res, _xy2ij

method_to_order = {"nearest": 0, "linear": 1, "cubic": 3, "quintic": 5, "slinear": 1, "pchip": 3, "splinef2d": 3}


def _get_dist_nodata_spread(order: int, dist_nodata_spread: Literal["half_order_up", "half_order_down"] | int) -> int:
    """
    Derive distance of nodata spreading based on interpolation order.

    :param order: Interpolation order.
    :param dist_nodata_spread: Spreading distance of nodata, either half-order rounded up (default), rounded down, or
        fixed integer.
    """

    if dist_nodata_spread == "half_order_up":
        dist_nodata_spread = int(np.ceil(order / 2))
    elif dist_nodata_spread == "half_order_down":
        dist_nodata_spread = int(np.floor(order / 2))

    return dist_nodata_spread


def _interpn_interpolator(
    points: tuple[NDArrayNum, NDArrayNum],
    values: NDArrayNum,
    fill_value: Number = np.nan,
    bounds_error: bool = False,
    dist_nodata_spread: Literal["half_order_up", "half_order_down"] | int = "half_order_up",
    method: Literal["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"] = "linear",
) -> Callable[[tuple[NDArrayNum, NDArrayNum]], NDArrayNum]:
    """
    Create SciPy interpolator with nodata spreading. Default is spreading at distance of half the method order
    rounded up (i.e., linear spreads 1 nodata in each direction, cubic spreads 2, quintic 3).

    Gives the exact same result as scipy.interpolate.interpn, and allows interpolator to be re-used if required (
    for speed).
    In practice, returns either a NaN-modified RegularGridInterpolator or a NaN-modified RectBivariateSpline object,
    both expecting a tuple of X/Y coordinates to be evaluated.

    For input arguments, see scipy.interpolate.RegularGridInterpolator.
    For additional argument "dist_nodata_spread", see description of Raster.interp_points.

    Adapted from:
    https://github.com/scipy/scipy/blob/44e4ebaac992fde33f04638b99629d23973cb9b2/scipy/interpolate/_rgi.py#L743.
    """

    # Derive distance to spread nodata to depending on method order
    order = method_to_order[method]
    d = _get_dist_nodata_spread(order=order, dist_nodata_spread=dist_nodata_spread)

    # We compute the nodata mask and dilate it to the distance to spread nodatas
    mask_nan = ~np.isfinite(values)
    if d != 0:
        new_mask = binary_dilation(mask_nan, iterations=d).astype("uint8")
    # Zero iterations has a different behaviour in binary_dilation than doing nothing, we want the original array
    else:
        new_mask = mask_nan.astype("uint8")

    # We create an interpolator for the nodata mask using nearest
    interp_mask = RegularGridInterpolator(points, new_mask, method="nearest", bounds_error=bounds_error, fill_value=1)

    # Most methods (cubic, quintic, etc) do not support NaNs and require an array full of valid values
    # We replace thus replace all NaN values by nearest neighbours to give surrounding values of the same order of
    # magnitude and minimize interpolation errors near NaNs (errors of 10e-2/e-5 relative to the values)
    # Elegant solution from: https://stackoverflow.com/questions/5551286/filling-gaps-in-a-numpy-array for a fast
    # nearest neighbour fill
    indices = distance_transform_edt(mask_nan, return_distances=False, return_indices=True)
    values = values[tuple(indices)]

    # For the RegularGridInterpolator
    if method in RegularGridInterpolator._ALL_METHODS:

        # We create the classic interpolator
        interp = RegularGridInterpolator(
            points, values, method=method, bounds_error=bounds_error, fill_value=fill_value
        )

        # We create a new interpolator callable that propagates nodata as defined above
        def regulargrid_interpolator_with_nan(xi: tuple[NDArrayNum, NDArrayNum]) -> NDArrayNum:

            # Get results
            results = interp(xi)
            # Get invalids
            invalids = interp_mask(xi)
            results[invalids.astype(bool)] = np.nan

            return results

        return regulargrid_interpolator_with_nan

    # For the RectBivariateSpline
    else:

        # The coordinates must be in ascending order, which requires flipping the array too (more costly)
        interp = RectBivariateSpline(np.flip(points[0]), points[1], np.flip(values[:], axis=0))

        # We create a new interpolator callable that propagates nodata as defined above, and supports fill_value
        def rectbivariate_interpolator_with_fillvalue(xi: tuple[NDArrayNum, NDArrayNum]) -> NDArrayNum:

            # Get invalids
            invalids = interp_mask(xi)

            # RectBivariateSpline doesn't support fill_value, so we need to wrap here to add them
            xi_arr = np.array(xi).T
            xi_shape = xi_arr.shape
            xi_arr = xi_arr.reshape(-1, xi_arr.shape[-1])
            idx_valid = np.all(
                (
                    points[0][-1] <= xi_arr[:, 0],
                    xi_arr[:, 0] <= points[0][0],
                    points[1][0] <= xi_arr[:, 1],
                    xi_arr[:, 1] <= points[1][-1],
                ),
                axis=0,
            )
            # Make a copy of values for RectBivariateSpline
            result = np.empty_like(xi_arr[:, 0])
            result[idx_valid] = interp.ev(xi_arr[idx_valid, 0], xi_arr[idx_valid, 1])
            result[np.logical_not(idx_valid)] = fill_value

            # Add back NaNs from dilated mask
            results = np.atleast_1d(result.reshape(xi_shape[:-1]))
            results[invalids.astype(bool)] = np.nan

            return results

        return rectbivariate_interpolator_with_fillvalue


def _map_coordinates_nodata_propag(
    values: NDArrayNum,
    indices: tuple[NDArrayNum, NDArrayNum],
    order: int,
    dist_nodata_spread: Literal["half_order_up", "half_order_down"] | int = "half_order_up",
    **kwargs: Any,
) -> NDArrayNum:
    """
    Perform map_coordinates with nodata spreading. Default is spreading at distance of half the method order rounded
    up (i.e., linear spreads 1 nodata in each direction, cubic spreads 2, quintic 3).

    For map_coordinates, only nearest and linear are used.

    For input arguments, see scipy.ndimage.map_coordinates.
    For additional argument "dist_nodata_spread", see description of Raster.interp_points.
    """

    # Derive distance of nodata spreading
    d = _get_dist_nodata_spread(order=order, dist_nodata_spread=dist_nodata_spread)

    # We compute the mask and dilate it to the distance to spread nodatas
    mask_nan = ~np.isfinite(values)
    if d != 0:
        new_mask = binary_dilation(mask_nan, iterations=d).astype("uint8")
    # Zero iterations has a different behaviour in binary_dilation than doing nothing, here we want the original array
    else:
        new_mask = mask_nan.astype("uint8")

    # We replace all NaN values by nearest neighbours to minimize interpolation errors near NaNs
    # Elegant solution from: https://stackoverflow.com/questions/5551286/filling-gaps-in-a-numpy-array
    ind = distance_transform_edt(mask_nan, return_distances=False, return_indices=True)
    values = values[tuple(ind)]

    # We interpolate the dilated array at the coordinates with nearest, and transform it back to a boolean to mask NaNs
    rmask = map_coordinates(new_mask, indices, order=0, cval=1, prefilter=False)

    # Interpolate at indices
    rpoints = map_coordinates(values, indices, order=order, **kwargs)

    # Set to NaNs based on spreading distance
    rpoints[rmask.astype(bool)] = np.nan

    return rpoints


@overload
def _interp_points(
    array: NDArrayNum,
    transform: rio.transform.Affine,
    area_or_point: Literal["Area", "Point"] | None,
    points: tuple[Number, Number] | tuple[NDArrayNum, NDArrayNum],
    method: Literal["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"] = "linear",
    dist_nodata_spread: Literal["half_order_up", "half_order_down"] | int = "half_order_up",
    shift_area_or_point: bool | None = None,
    force_scipy_function: Literal["map_coordinates", "interpn"] | None = None,
    *,
    return_interpolator: Literal[False] = False,
    **kwargs: Any,
) -> NDArrayNum: ...


@overload
def _interp_points(
    array: NDArrayNum,
    transform: rio.transform.Affine,
    area_or_point: Literal["Area", "Point"] | None,
    points: tuple[Number, Number] | tuple[NDArrayNum, NDArrayNum],
    method: Literal["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"] = "linear",
    dist_nodata_spread: Literal["half_order_up", "half_order_down"] | int = "half_order_up",
    shift_area_or_point: bool | None = None,
    force_scipy_function: Literal["map_coordinates", "interpn"] | None = None,
    *,
    return_interpolator: Literal[True],
    **kwargs: Any,
) -> Callable[[tuple[NDArrayNum, NDArrayNum]], NDArrayNum]: ...


@overload
def _interp_points(
    array: NDArrayNum,
    transform: rio.transform.Affine,
    area_or_point: Literal["Area", "Point"] | None,
    points: tuple[Number, Number] | tuple[NDArrayNum, NDArrayNum],
    method: Literal["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"] = "linear",
    dist_nodata_spread: Literal["half_order_up", "half_order_down"] | int = "half_order_up",
    shift_area_or_point: bool | None = None,
    force_scipy_function: Literal["map_coordinates", "interpn"] | None = None,
    *,
    return_interpolator: bool = False,
    **kwargs: Any,
) -> NDArrayNum | Callable[[tuple[NDArrayNum, NDArrayNum]], NDArrayNum]: ...


@profiler.profile("geoutils.interface.interpolate._interp_points", memprof=True)  # type: ignore
def _interp_points(
    array: NDArrayNum,
    transform: rio.transform.Affine,
    area_or_point: Literal["Area", "Point"] | None,
    points: tuple[Number, Number] | tuple[NDArrayNum, NDArrayNum] | None,
    method: Literal["nearest", "linear", "cubic", "quintic", "slinear", "pchip", "splinef2d"] = "linear",
    dist_nodata_spread: Literal["half_order_up", "half_order_down"] | int = "half_order_up",
    shift_area_or_point: bool | None = None,
    force_scipy_function: Literal["map_coordinates", "interpn"] | None = None,
    return_interpolator: bool = False,
    **kwargs: Any,
) -> NDArrayNum | Callable[[tuple[NDArrayNum, NDArrayNum]], NDArrayNum]:
    """See description of Raster.interp_points."""

    # If array is not a floating dtype (to support NaNs), convert dtype
    if not np.issubdtype(array.dtype, np.floating):
        array = array.astype(np.float32)
    shape: tuple[int, int] = array.shape[0:2]  # type: ignore

    # TODO: Add check about None for "points" depending on "return_interpolator"
    if not return_interpolator:
        if points is None:
            raise ValueError("Input 'points' cannot be None if 'return_interpolator' is False.")
        x, y = points
        i, j = _xy2ij(x, y, transform=transform, area_or_point=area_or_point, shift_area_or_point=shift_area_or_point)

        ind_invalid = np.vectorize(
            lambda k1, k2: _outside_image(
                k1, k2, transform=transform, area_or_point=area_or_point, shape=shape, index=True
            )
        )(j, i)

    # If the raster is on an equal grid, use scipy.ndimage.map_coordinates
    force_map_coords = force_scipy_function is not None and force_scipy_function == "map_coordinates"
    force_interpn = force_scipy_function is not None and force_scipy_function == "interpn"

    # Map method name to spline order in map_coordinates, and use only is method compatible
    method_to_order_mapcoords = {"nearest": 0, "linear": 1}
    mapcoords_supported = method in method_to_order_mapcoords.keys()

    res = _res(transform)
    if (res[0] == res[1] or force_map_coords) and not force_interpn and mapcoords_supported and not return_interpolator:

        # Convert method name into order
        order = method_to_order_mapcoords[method]

        # Remove default spline pre-filtering that is activated by default
        if "prefilter" not in kwargs.keys():
            kwargs.update({"prefilter": False})
        # Change default constant value to NaN for interpolation outside the image bounds
        if "cval" not in kwargs.keys():
            kwargs.update({"cval": np.nan})

        # Use map coordinates with nodata propagation
        rpoints = _map_coordinates_nodata_propag(
            values=array, indices=(i, j), order=order, dist_nodata_spread=dist_nodata_spread, **kwargs
        )

    # Otherwise, use scipy.interpolate.interpn
    else:
        # Get lower-left corner coordinates
        xycoords = _coords(
            transform=transform,
            shape=shape,
            area_or_point=area_or_point,
            grid=False,
            shift_area_or_point=shift_area_or_point,
        )

        # Let interpolation outside the bounds not raise any error by default
        if "bounds_error" not in kwargs.keys():
            kwargs.update({"bounds_error": False})
        # Return NaN outside image bounds
        if "fill_value" not in kwargs.keys():
            kwargs.update({"fill_value": np.nan})

        # Using direct coordinates, Y is the first axis, and we need to flip it
        interpolator = _interpn_interpolator(
            points=(np.flip(xycoords[1], axis=0), xycoords[0]),
            values=array,
            method=method,
            dist_nodata_spread=dist_nodata_spread,
            bounds_error=kwargs["bounds_error"],
            fill_value=kwargs["fill_value"],
        )
        if return_interpolator:
            return interpolator
        else:
            rpoints = interpolator((y, x))  # type: ignore

    rpoints = np.array(np.atleast_1d(rpoints), dtype=np.float32)
    rpoints[np.array(ind_invalid)] = np.nan

    return rpoints



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for PointCloud class."""

from __future__ import annotations

import logging
import os.path
import pathlib
import warnings
from typing import TYPE_CHECKING, Any, Callable, Iterable, Literal, TypeVar, overload

import geopandas as gpd
import numpy as np
import pandas as pd
from pyproj import CRS
from rasterio.coords import BoundingBox
from rasterio.transform import from_origin
from shapely.geometry.base import BaseGeometry

import geoutils as gu
from geoutils import profiler
from geoutils._misc import import_optional
from geoutils._typing import ArrayLike, DTypeLike, NDArrayBool, NDArrayNum, Number
from geoutils.interface.gridding import _grid_pointcloud
from geoutils.raster.georeferencing import _coords
from geoutils.stats.sampling import subsample_array
from geoutils.stats.stats import _statistics

if TYPE_CHECKING:
    import matplotlib

# List of NumPy "array" functions that are handled.
# Note: all universal function are supported: https://numpy.org/doc/stable/reference/ufuncs.html
# Array functions include: NaN math and stats, classic math and stats, logical, sorting/counting:
_HANDLED_FUNCTIONS_1NIN = (
    # NaN math: https://numpy.org/doc/stable/reference/routines.math.html
    # and NaN stats: https://numpy.org/doc/stable/reference/routines.statistics.html
    [
        "nansum",
        "nanmax",
        "nanmin",
        "nanargmax",
        "nanargmin",
        "nanmean",
        "nanmedian",
        "nanpercentile",
        "nanvar",
        "nanstd",
        "nanprod",
        "nancumsum",
        "nancumprod",
        "nanquantile",
    ]
    # Classic math and stats (same links as above)
    + [
        "sum",
        "amax",
        "amin",
        "max",
        "min",
        "argmax",
        "argmin",
        "mean",
        "median",
        "percentile",
        "var",
        "std",
        "prod",
        "cumsum",
        "cumprod",
        "quantile",
        "abs",
        "absolute",
    ]
    # Sorting, searching and counting: https://numpy.org/doc/stable/reference/routines.sort.html
    + ["sort", "count_nonzero", "unique"]
    # Logic functions: https://numpy.org/doc/stable/reference/routines.logic.html
    + ["all", "any", "isfinite", "isinf", "isnan", "logical_not"]
)

_HANDLED_FUNCTIONS_2NIN = [
    "logical_and",
    "logical_or",
    "logical_xor",
    "allclose",
    "isclose",
    "array_equal",
    "array_equiv",
    "greater",
    "greater_equal",
    "less",
    "less_equal",
    "equal",
    "not_equal",
]
handled_array_funcs = _HANDLED_FUNCTIONS_1NIN + _HANDLED_FUNCTIONS_2NIN

# This is a generic Vector-type (if subclasses are made, this will change appropriately)
PointCloudType = TypeVar("PointCloudType", bound="PointCloud")


def _load_laspy_data(filename: str, columns: list[str]) -> gpd.GeoDataFrame:
    """Load point cloud data from LAS/LAZ/COPC file as a geodataframe."""

    laspy = import_optional("laspy")

    # Read file
    las = laspy.read(filename)

    # Get data from main Z column and other requested columns
    columns_no_z = [c for c in columns if c != "Z"]
    data = np.vstack([las.z] + [las[n] for n in columns_no_z]).T
    column_names = ["Z"] + columns_no_z

    # Build geodataframe
    gdf = gpd.GeoDataFrame(
        geometry=gpd.points_from_xy(x=las.x, y=las.y, crs=las.header.parse_crs(prefer_wkt=False)),
        data=data,
        columns=column_names,
    )

    return gdf


def _load_laspy_metadata(
    filename: str,
) -> tuple[CRS, int, BoundingBox, pd.Index]:
    """Load point cloud metadata from LAS/LAZ/COPC file."""

    laspy = import_optional("laspy")

    with laspy.open(filename) as f:

        # Parse CRS, point count and bounds
        crs = f.header.parse_crs(prefer_wkt=False)
        nb_points = f.header.point_count
        bounds = BoundingBox(left=f.header.x_min, right=f.header.x_max, bottom=f.header.y_min, top=f.header.y_max)

        # Parse column names as pandas index, removing X/Y that will be transformed into a geometry
        columns_names = list(f.header.point_format.dimension_names)
        columns_names = [c for c in columns_names if c not in ["X", "Y"]]
        columns_names = pd.Index(columns_names)

    return crs, nb_points, bounds, columns_names


def _write_laspy(
    filename: str | pathlib.Path,
    pc: gpd.GeoDataFrame,
    data_column: str | None,
    version: Any = None,
    point_format: Any = None,
    offsets: tuple[float, float, float] = None,
    scales: tuple[float, float, float] = None,
    **kwargs: Any,
) -> None:
    """Write a point cloud geodataframe to a LAS/LAZ/COPC file."""

    laspy = import_optional("laspy")

    # Initiate header with user arguments
    header = laspy.LasHeader(version=version, point_format=point_format)
    if scales is not None:
        header.scales = np.array(scales)
    if offsets is not None:
        header.offsets = np.array(offsets)
    for k, v in kwargs.items():
        setattr(header, k, v)

    # Adding extra dimensions for auxiliary variables
    aux_columns = [c for c in pc.columns if c not in [data_column, "geometry"]]
    for c in aux_columns:
        header.add_extra_dim(laspy.ExtraBytesParams(name=c, type=pc[c].dtype))

    las = laspy.LasData(header)

    # The las x,y,z will be automatically scaled into X,Y,Z based on the header scales
    las.x = pc.geometry.x.values
    las.y = pc.geometry.y.values
    if data_column is not None:
        las.z = pc[data_column].values
    else:
        las.z = pc.geometry.z.values

    # Add auxiliary columns
    for c in aux_columns:
        setattr(las, c, pc[c].values)

    las.write(filename)

    return


def _cast_numeric_array_pointcloud(
    pc: PointCloudType, other: PointCloudType | NDArrayNum | Number, operation_name: str
) -> NDArrayNum | Number:
    """
    Cast a point cloud and another point cloud or array or number to arrays with proper metadata, or raise an error
    message.

    :param pc: Pointcloud.
    :param other: Point cloud or array or number.
    :param operation_name: Name of operation to raise in the error message.
    """

    # Check first input is a point cloud
    if not isinstance(pc, PointCloud):
        raise ValueError("Developer error: Only a point cloud should be passed as first argument to this function.")

    # Check that other is of correct type
    # If not, a NotImplementedError should be raised, in case other's class has a method implemented.
    # See https://docs.python.org/3/reference/datamodel.html#emulating-numeric-types
    if not isinstance(other, (PointCloud, np.ndarray, float, int, np.floating, np.integer)):
        raise NotImplementedError(
            f"Operation between an object of type {type(other)} and a point cloud impossible. Must be a point cloud, "
            f"np.ndarray or single number."
        )

    # If other is a point cloud
    if isinstance(other, PointCloud):

        other_data = other.data
        # Check that both point clouds have the same shape and georeferences
        if pc.georeferenced_coords_equal(other):  # type: ignore
            pass
        else:
            raise ValueError(
                "Both point clouds must have the same points X/Y coordinates and CRS for " + operation_name + "."
            )

    # If other is an array
    elif isinstance(other, np.ndarray):

        other_data = other.squeeze()
        if other.squeeze().ndim == 1 and other.squeeze().shape[0] == pc.point_count:
            pass
        else:
            raise ValueError(
                "The array must be 1-dimensional with the same number of points as the point cloud for "
                + operation_name
                + "."
            )

    else:
        other_data = other  # type: ignore

    return other_data


class PointCloud(gu.Vector):  # type: ignore[misc]
    """
    The georeferenced point cloud.

    A point cloud is a vector of 2D point geometries associated to numeric values from a main data column, and can
    also contain auxiliary data columns.

     Main attributes:
        ds: :class:`geopandas.GeoDataFrame`
            Geodataframe of the point cloud.
        data_column: str
            Name of point cloud data column.
        crs: :class:`pyproj.crs.CRS`
            Coordinate reference system of the point cloud.
        bounds: :class:`rio.coords.BoundingBox`
            Coordinate bounds of the point cloud.


    All other attributes are derivatives of those attributes, or read from the file on disk.
    See the API for more details.
    """

    @profiler.profile("geoutils.pointcloud.pointcloud.__init__", memprof=True)  # type: ignore
    def __init__(
        self,
        filename_or_dataset: str | pathlib.Path | gpd.GeoDataFrame | gpd.GeoSeries | BaseGeometry,
        data_column: str | None = None,
    ):
        """
        Instantiate a point cloud from either a data column name and a vector (filename, GeoPandas dataframe or series,
        or a Shapely geometry), or only with a point cloud file type.

        :param filename_or_dataset: Path to vector file, or GeoPandas dataframe or series, or Shapely geometry.
        :param data_column: Name of main data column defining the point cloud (not required for LAS/LAZ formats).
        """

        self._ds: gpd.GeoDataFrame | None = None
        self._name: str | None = None
        self._crs: CRS | None = None
        self._data_column: str | None = None
        self._bounds: BoundingBox
        self._data: NDArrayNum
        self._nb_points: int
        self.__nongeo_columns: pd.Index

        # If PointCloud is passed, simply point back to PointCloud
        if isinstance(filename_or_dataset, PointCloud):
            for key in filename_or_dataset.__dict__:
                setattr(self, key, filename_or_dataset.__dict__[key])
            return
        # For filename, rely on parent Vector class or LAS file reader
        else:
            if isinstance(filename_or_dataset, (str, pathlib.Path)) and os.path.splitext(filename_or_dataset)[-1] in [
                ".las",
                ".laz",
            ]:

                # No need to pass a data column for LAS/LAZ file, as Z is the logical default
                if data_column is None:
                    data_column = "Z"
                # Load only metadata, and not the data
                fn = filename_or_dataset if isinstance(filename_or_dataset, str) else filename_or_dataset.name
                crs, nb_points, bounds, columns = _load_laspy_metadata(fn)
                self._name = fn
                self._crs = crs
                self._nb_points = nb_points
                self.__nongeo_columns = columns
                self._bounds = bounds
                self._ds = None
            # Check on filename are done with Vector.__init__
            else:
                super().__init__(filename_or_dataset)
                # Verify that the vector can be cast as a point cloud
                if not all(p == "Point" for p in self.ds.geom_type):
                    raise ValueError(
                        "This vector file contains non-point geometries, " "cannot be instantiated as a point cloud."
                    )

        # Set data column name based on user input
        self.set_data_column(new_data_column=data_column)

    # TODO: Could also move to Vector directly?
    ##############################################
    # OVERRIDDEN VECTOR METHODS TO SUPPORT LOADING
    ##############################################

    @property
    def ds(self) -> gpd.GeoDataFrame:
        """Geodataframe of the point cloud."""
        # We need to override the Vector method to introduce the is_loaded dynamic for LAS files
        if not self.is_loaded:
            self.load()
        return self._ds  # type: ignore

    @ds.setter
    def ds(self, new_ds: gpd.GeoDataFrame | gpd.GeoSeries) -> None:
        """Set a new geodataframe for the point cloud."""
        # We need to override the setter Vector method because we have overridden the property method
        # (even if the code below is the same)
        if isinstance(new_ds, gpd.GeoDataFrame):
            self._ds = new_ds
        elif isinstance(new_ds, gpd.GeoSeries):
            self._ds = gpd.GeoDataFrame(geometry=new_ds)
        else:
            raise ValueError("The dataset of a vector must be set with a GeoSeries or a GeoDataFrame.")

    @property
    def crs(self) -> CRS:
        """Coordinate reference system of the vector."""

        # Overriding method in Vector in case dataset is not loaded
        if self.is_loaded:
            return super().crs
        # Return CRS on disk
        else:
            return self._crs

    @property
    def bounds(self) -> BoundingBox:
        # Overriding method in Vector in case dataset is not loaded
        if self.is_loaded:
            return super().bounds
        # Return bounds on disk
        else:
            return self._bounds

    def columns(self) -> pd.Index:
        # Overriding method in Vector in case dataset is not loaded
        if self.is_loaded:
            return super().columns
        # Return columns on disk (adding a placeholder geometry to replace X/Y)
        else:
            return pd.Index(list(self._nongeo_columns) + ["geometry"])

    #####################################
    # NEW METHODS SPECIFIC TO POINT CLOUD
    #####################################

    @property
    def _has_z(self) -> bool:
        """Whether the point geometries all have a Z coordinate or not."""

        return all(p.has_z for p in self.ds.geometry) if len(self.ds.geometry) > 0 else False

    @property
    def data(self) -> NDArrayNum:
        """
        Data of the point cloud.

        Points to either the Z axis of the point geometries, or the associated data column of the geodataframe.
        """
        # Triggers the loading mechanism through self.ds
        if self.data_column is not None:
            return self.ds[self.data_column].values
        else:
            return self.geometry.z.values

    @data.setter
    def data(self, new_data: NDArrayNum) -> None:
        """Set new data for the point cloud."""

        if self.data_column is not None:
            self.ds[self.data_column] = new_data
        else:
            self.ds.geometry = gpd.points_from_xy(x=self.geometry.x, y=self.geometry.y, z=new_data, crs=self.crs)

    @property
    def _nongeo_columns(self) -> pd.Index:
        """Columns of the point cloud excluding the column of 2D point geometries."""
        # Overriding method in Vector
        if self.is_loaded:
            nongeo_columns = super().columns
            nongeo_columns = nongeo_columns[nongeo_columns != "geometry"]
            return nongeo_columns
        else:
            return self.__nongeo_columns

    @property
    def data_column(self) -> str | None:
        """
        Name of data column of the point cloud.

        Can be None if point geometries are 3D.
        """
        return self._data_column

    @data_column.setter
    def data_column(self, new_data_column: str) -> None:
        self.set_data_column(new_data_column=new_data_column)

    def set_data_column(self, new_data_column: str | None) -> None:
        """Set new column as point cloud data column."""

        # If point geometries are 3D, only for loaded data (otherwise _has_z would load data)
        if self.is_loaded:
            if self._has_z:
                if new_data_column is None:
                    self._data_column = None
                    return
                else:
                    warnings.warn(
                        f"Overriding 3D points with with data column '{new_data_column}'. Set data_column "
                        f"to None to use the 3D point geometries instead."
                    )

        # If point geometries are 2D and the data column is undefined
        if new_data_column is None:
            raise ValueError("A data column name must be passed for a point cloud with 2D point geometries.")

        # If 2D and data column is defined, check that it exists
        if new_data_column not in self._nongeo_columns:
            raise ValueError(
                f"Data column {new_data_column} not found among columns. Available columns "
                f"are: {', '.join(self._nongeo_columns)}."
            )

        # Set data column name
        self._data_column = new_data_column

    @property
    def is_loaded(self) -> bool:
        """Whether the point cloud data is loaded."""
        return self._ds is not None

    @property
    def point_count(self) -> int:
        """Number of points in the point cloud."""
        # New method for point cloud
        if self.is_loaded:
            return len(self.ds)
        else:
            return self._nb_points

    @property
    def is_mask(self) -> bool:
        """Whether the point cloud mask is a mask (boolean type)."""

        return np.dtype(self.data.dtype) == np.bool_

    def load(self, columns: Literal["all", "main"] | list[str] = "main") -> None:
        """
        Load point cloud from disk (only supported for LAS files).

        :param columns: Columns to load. Defaults to main data column only.
        """

        if self.is_loaded:
            raise ValueError("Data are already loaded.")

        if self.name is None:
            raise AttributeError(
                "Cannot load as filename is not set anymore. Did you manually update the filename attribute?"
            )

        if columns == "all":
            columns_to_load = self._nongeo_columns
        elif columns == "main":
            columns_to_load = [self.data_column]
        else:
            columns_to_load = columns

        ds = _load_laspy_data(filename=self.name, columns=columns_to_load)
        self._ds = ds

    @overload
    def astype(
        self: PointCloud, dtype: DTypeLike, convert_coords: bool = False, *, inplace: Literal[False] = False
    ) -> PointCloud: ...

    @overload
    def astype(self: PointCloud, dtype: DTypeLike, convert_coords: bool = False, *, inplace: Literal[True]) -> None: ...

    @overload
    def astype(
        self: PointCloud, dtype: DTypeLike, convert_coords: bool = False, *, inplace: bool = False
    ) -> PointCloud | None: ...

    def astype(
        self: PointCloud, dtype: DTypeLike, convert_coords: bool = False, inplace: bool = False
    ) -> PointCloud | None:
        """
        Convert data type of the point cloud data column.

        :param dtype: Any numpy dtype or string accepted by numpy.astype.
        :param convert_coords: Whether to convert the data type of coordinates values as well.
        :param inplace: Whether to modify the point cloud in-place.

        :returns: Point cloud with updated dtype (or None if inplace).
        """

        out_data = self.data.astype(dtype)

        if inplace:
            self._data = out_data  # type: ignore
            if convert_coords:
                self.ds.geometry.x = self.ds.geometry.x.values.astype(dtype)
                self.ds.geometry.y = self.ds.geometry.y.values.astype(dtype)
            return None
        else:
            if convert_coords:
                x = self.ds.geometry.x.values.astype(dtype)
                y = self.ds.geometry.y.values.astype(dtype)
                return gu.PointCloud.from_xyz(x=x, y=y, z=out_data, crs=self.crs, data_column=self.data_column)
            else:
                return self.copy(new_array=out_data)

    def copy(self: PointCloud, new_array: NDArrayNum | NDArrayBool | None = None) -> PointCloud:
        """
        Copy the point cloud in-memory.

        :param new_array: New data array to use in the copied point cloud's data column.

        :return: Copy of the point cloud.
        """

        # Define new array
        if new_array is not None:
            if not isinstance(new_array, np.ndarray):
                raise ValueError("New data must be an array.")
            new_array = new_array.squeeze()
            if not (new_array.ndim == 1 and new_array.shape[0] == self.point_count):
                raise ValueError(
                    "New data array must be 1-dimensional with the same number of points as the point "
                    "cloud being copied."
                )
            data = new_array
        else:
            data = self.data.copy()

        # Send to from_xyz
        cp = self.from_xyz(
            x=self.geometry.x.values,
            y=self.geometry.y.values,
            z=data,
            crs=self.crs,
            data_column=self.data_column,
            use_z=self._has_z and self.data_column is None,
        )

        return cp

    def to_las(
        self,
        filename: str | pathlib.Path,
        version: Any = None,
        point_format: Any = None,
        offsets: tuple[float, float, float] = None,
        scales: tuple[float, float, float] = None,
        **kwargs: Any,
    ) -> None:
        """
        Write the point cloud to LAS/LAZ/COPC file.

        :param filename: Name of output file.
        :param version: LAS/LAZ/COPC version.
        :param point_format: Point format.
        :param offsets: Offsets for X/Y/Z.
        :param scales: Scales for X/Y/Z.
        :param kwargs: Other keyword arguments to set the LAS file header (e.g., "offsets", "scales").
        """

        _write_laspy(
            filename=filename,
            pc=self.ds,
            data_column=self.data_column,
            version=version,
            point_format=point_format,
            offsets=offsets,
            scales=scales,
            **kwargs,
        )

    @classmethod
    def from_xyz(
        cls, x: ArrayLike, y: ArrayLike, z: ArrayLike, crs: CRS, data_column: str | None = None, use_z: bool = False
    ) -> PointCloud:
        """
        Create point cloud from three 1D array-like coordinates for X/Y/Z.

        Note that this is the most modular method to create a point cloud, as it allows to specify different data
        types for the different coordinates or columns.

        :param x: X coordinates of point cloud.
        :param y: Y coordinates of point cloud.
        :param z: Z values of point cloud.
        :param crs: Coordinate reference system.
        :param data_column: Data column name to associate to 2D point geometries (defaults to "z" if none is passed).
        :param use_z: Use 3D point geometries with Z coordinates instead of a data column.

        :return Point cloud.
        """

        # Build geodataframe
        if not use_z:
            data_column = data_column if data_column is not None else "z"
            gdf = gpd.GeoDataFrame(
                geometry=gpd.points_from_xy(x=np.atleast_1d(x), y=np.atleast_1d(y), crs=crs),
                data={data_column: np.atleast_1d(z)},
            )
        else:
            data_column = None
            gdf = gpd.GeoDataFrame(
                geometry=gpd.points_from_xy(x=np.atleast_1d(x), y=np.atleast_1d(y), z=np.atleast_1d(z), crs=crs),
            )

        # If the data was transformed into boolean, re-initialize as a Mask subclass
        # Typing: we can specify this behaviour in @overload once we add the NumPy plugin of MyPy
        if np.atleast_1d(z)[0].dtype == bool:
            return PointCloud(filename_or_dataset=gdf, data_column=data_column)  # type: ignore
        # Otherwise, keep as a given PointCloudType subclass
        else:
            return cls(filename_or_dataset=gdf, data_column=data_column)

    @classmethod
    def from_array(cls, data: NDArrayNum, crs: CRS, data_column: str | None = None, use_z: bool = False) -> PointCloud:
        """
        Create point cloud from a 3 x N or N x 3 array of X coordinates, Y coordinates and Z values.

        :param data: Point cloud coordinates and data values as 3 x N or N x 3 array.
        :param crs: Coordinate reference system of point cloud.
        :param data_column: Data column of point cloud.

        :return Point cloud.
        """

        # Check shape
        if data.ndim != 2 or (data.shape[0] != 3 and data.shape[1] != 3):
            raise ValueError("Array must be of shape 3xN or Nx3.")

        # Make the first axis the one with size 3
        if data.shape[0] != 3:
            data = data.T

        return cls.from_xyz(x=data[0, :], y=data[1, :], z=data[2, :], crs=crs, data_column=data_column, use_z=use_z)

    @classmethod
    def from_tuples(
        cls,
        tuples_xyz: Iterable[tuple[Number, Number, Number]],
        crs: CRS,
        data_column: str | None = None,
        use_z: bool = False,
    ) -> PointCloud:
        """
        Create point cloud from an iterable of 3-tuples (X coordinate, Y coordinate, Z value).

        :param tuples_xyz: Point cloud coordinates and data as an iterable of 3-tuples.
        :param crs: Coordinate reference system of point cloud.
        :param data_column: Data column of point cloud.

        :return Point cloud.
        """

        return cls.from_array(np.array(tuples_xyz), crs=crs, data_column=data_column, use_z=use_z)

    def to_xyz(self) -> tuple[NDArrayNum, NDArrayNum, NDArrayNum]:
        """Convert point cloud to three 1D arrays of coordinates for X/Y/Z."""

        return self.geometry.x.values, self.geometry.y.values, self.data

    def to_array(self) -> NDArrayNum:
        """Convert point cloud to a 3 x N array of X coordinates, Y coordinates and Z values."""

        return np.stack((self.geometry.x.values, self.geometry.y.values, self.data), axis=0)

    def to_tuples(self) -> Iterable[tuple[Number, Number, Number]]:
        """Convert point cloud to a list of 3-tuples (X coordinate, Y coordinate, Z value)."""

        return list(zip(self.geometry.x.values, self.geometry.y.values, self.data))

    def __getitem__(self, index: PointCloud | NDArrayBool | Any) -> PointCloud | Any:
        """
        Index the point cloud.

        In addition to all index types supported by GeoPandas, also supports a point cloud mask of same georeferencing.
        """

        # If input is mask with the same shape and georeferencing
        if isinstance(index, PointCloud) or (isinstance(index, np.ndarray) and len(index) == self.point_count):
            _cast_numeric_array_pointcloud(self, index, operation_name="an indexing operation")  # type: ignore
            if isinstance(index, PointCloud):
                ind = index.data
            else:
                ind = index  # type: ignore
            ind = ind.astype(bool)  # In case the 3D Z column was used, it can only be stored as floating

            return PointCloud(self.ds.loc[ind], data_column=self.data_column)

        # Otherwise, use index and leave it to GeoPandas
        else:
            ind = index  # type: ignore
            return super().__getitem__(ind)

    def __setitem__(self, index: Any, assign: Any) -> None:
        """
        Perform index assignment on the point cloud.
        """

        # If input is mask with the same shape and georeferencing
        if isinstance(index, PointCloud) or (isinstance(index, np.ndarray) and len(index) == self.point_count):
            _cast_numeric_array_pointcloud(self, index, operation_name="an indexing operation")  # type: ignore
            # Get index
            if isinstance(index, PointCloud):
                ind = index.data
            else:
                ind = index  # type: ignore
            ind = ind.astype(bool)  # In case the 3D Z column was used, it can only be stored as floating
            # Assign
            if self._has_z:
                print(ind)
                new_geo = gpd.points_from_xy(
                    x=self.geometry.x.values[ind], y=self.geometry.y.values[ind], z=assign, crs=self.crs
                )
                self.ds.loc[ind, "geometry"] = new_geo
            else:
                print(self.data_column)
                self.ds.loc[ind, [self.data_column]] = assign

        else:
            # Let the vector class do the job
            super().__setitem__(index, assign)

        return None

    def __array_ufunc__(
        self,
        ufunc: Callable[[NDArrayNum | tuple[NDArrayNum, NDArrayNum]], NDArrayNum | tuple[NDArrayNum, NDArrayNum]],
        method: str,
        *inputs: tuple[PointCloud]
        | tuple[PointCloud, PointCloud]
        | tuple[NDArrayNum, PointCloud]
        | tuple[PointCloud, NDArrayNum],
        **kwargs: Any,
    ) -> PointCloud | tuple[PointCloud, PointCloud]:
        """
        Method to cast NumPy universal functions directly on PointCloud classes, by passing to the masked array.
        This function basically applies the ufunc (with its method and kwargs) to .data, and rebuilds the PointCloud
        from self.__class__. The cases separate the number of input nin and output nout, to properly feed .data and
        return PointCloud objects.
        See more details in NumPy doc, e.g., https://numpy.org/doc/stable/user/basics.dispatch.html#basics-dispatch.
        """

        # In addition to running ufuncs, this function takes over arithmetic operations (__add__, __multiply__, etc...)
        # when the first input provided is a NumPy array and second input a PointCloud.
        final_ufunc = getattr(ufunc, method)

        # If the universal function takes only one input
        if ufunc.nin == 1:
            # If the universal function has only one output
            if ufunc.nout == 1:
                return self.copy(new_array=final_ufunc(inputs[0].data, **kwargs))  # type: ignore

            # If the universal function has two outputs (Note: no ufunc exists that has three outputs or more)
            else:
                output = final_ufunc(inputs[0].data, **kwargs)  # type: ignore
                return self.copy(new_array=output[0]), self.copy(new_array=output[1])

        # If the universal function takes two inputs (Note: no ufunc exists that has three inputs or more)
        else:

            # Check the casting between Point cloud and array inputs, and return error messages if not consistent

            # Raise errors if necessary
            if isinstance(inputs[0], PointCloud):
                pc = inputs[0]
                other = inputs[1]
            else:
                pc = inputs[1]  # type: ignore
                other = inputs[0]
            _ = _cast_numeric_array_pointcloud(pc, other, "an arithmetic operation")  # type: ignore

            # Get data depending on argument order
            if isinstance(inputs[0], PointCloud):
                first_arg = inputs[0].data
            else:
                first_arg = inputs[0]

            if isinstance(inputs[1], PointCloud):
                second_arg = inputs[1].data
            else:
                second_arg = inputs[1]

            # For one output
            if ufunc.nout == 1:
                return self.copy(new_array=final_ufunc(first_arg, second_arg, **kwargs))

            # If the universal function has two outputs (Note: no ufunc exists that has three outputs or more)
            else:
                output = final_ufunc(first_arg, second_arg, **kwargs)  # type: ignore
                return self.copy(new_array=output[0]), self.copy(new_array=output[1])

    def __array_function__(
        self, func: Callable[[NDArrayNum, Any], Any], types: tuple[type], args: Any, kwargs: Any
    ) -> Any:
        """
        Method to cast NumPy array function directly on a Point cloud object by applying it to the masked array.
        A limited number of function is supported, listed in point cloud.handled_array_funcs.
        """

        # If function is not implemented
        if func.__name__ not in _HANDLED_FUNCTIONS_1NIN + _HANDLED_FUNCTIONS_2NIN:
            return NotImplemented

        # For subclassing
        if not all(issubclass(t, self.__class__) for t in types):
            return NotImplemented

        # Get first argument
        first_arg = args[0].data

        # Separate one and two input functions
        if func.__name__ in _HANDLED_FUNCTIONS_1NIN:
            outputs = func(first_arg, *args[1:], **kwargs)  # type: ignore
        # Two input functions require casting
        else:
            # Check the casting between point cloud and array inputs, and return error messages if not consistent
            if isinstance(args[0], PointCloud):
                pc = args[0]
                other = args[1]
            else:
                pc = args[1]
                other = args[0]
            _ = _cast_numeric_array_pointcloud(pc, other, operation_name="an arithmetic operation")
            second_arg = args[1].data
            outputs = func(first_arg, second_arg, *args[2:], **kwargs)  # type: ignore

        # Below, we recast to PointCloud if the shape was preserved, otherwise return an array
        # First, if there are several outputs in a tuple which are arrays
        if isinstance(outputs, tuple) and isinstance(outputs[0], np.ndarray):
            if all(output.shape == args[0].data.shape for output in outputs):
                return tuple(self.copy(new_array=output) for output in outputs)
            else:
                return outputs
        # Second, if there is a single output which is an array
        elif isinstance(outputs, np.ndarray):
            if outputs.shape == args[0].data.shape:
                return self.copy(new_array=outputs)
            else:
                return outputs
        # Else, return outputs directly
        else:
            return outputs

    def plot(
        self,
        column: str | None = None,
        ref_crs: gu.Raster | gu.Vector | gpd.GeoDataFrame | str | CRS | int | None = None,
        cmap: matplotlib.colors.Colormap | str | None = None,
        vmin: float | int | None = None,
        vmax: float | int | None = None,
        alpha: float | int | None = None,
        cbar_title: str | None = None,
        add_cbar: bool = True,
        ax: matplotlib.axes.Axes | Literal["new"] | None = None,
        return_axes: bool = False,
        savefig_fname: str | None = None,
        **kwargs: Any,
    ) -> None | tuple[matplotlib.axes.Axes, matplotlib.colors.Colormap]:
        r"""
        Plot the point cloud.

        This method is a wrapper to geopandas.GeoDataFrame.plot. Any \*\*kwargs which
        you give this method will be passed to it.

        :param column: Column to plot. Default is the data column of the point cloud.
        :param ref_crs: Coordinate reference system to match when plotting.
        :param cmap: Colormap to use. Default is plt.rcParams['image.cmap'].
        :param vmin: Colorbar minimum value. Default is data min.
        :param vmax: Colorbar maximum value. Default is data max.
        :param alpha: Transparency of raster and colorbar.
        :param cbar_title: Colorbar label. Default is None.
        :param add_cbar: Set to True to display a colorbar. Default is True if a "column" argument is passed.
        :param ax: A figure ax to be used for plotting. If None, will plot on current axes. If "new",
            will create a new axis.
        :param return_axes: Whether to return axes.
        :param savefig_fname: Path to quick save the output figure (previously created if an ax is give, new if not)
            with a default DPI, no transparency and no metadata. Use `plt.savefig()` to specify other save
            parameters or after other customizations. Warning: `plt.close()` or `plt.show()` still needs to be called
            to close the figure.

        :returns: None, or (ax, caxes) if return_axes is True
        """

        matplotlib = import_optional("matplotlib")
        import matplotlib.pyplot as plt
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        # Ensure that the vector is in the same crs as a reference
        if isinstance(ref_crs, (gu.Raster, gu.Vector, gpd.GeoDataFrame, str)):
            vect_reproj = self.reproject(ref=ref_crs)
        elif isinstance(ref_crs, (CRS, int)):
            vect_reproj = self.reproject(crs=ref_crs)
        else:
            vect_reproj = self

        if column is None:
            column = self.data_column

        # Create axes, or get current ones by default (like in matplotlib)
        if ax is None:
            ax0 = plt.gca()
        elif isinstance(ax, str) and ax.lower() == "new":
            _, ax0 = plt.subplots()
        elif isinstance(ax, matplotlib.axes.Axes):
            ax0 = ax
        else:
            raise ValueError("ax must be a matplotlib.axes.Axes instance, 'new' or None.")

        # Set add_cbar depending on column argument
        if add_cbar:
            add_cbar = True
        else:
            add_cbar = False

        # Update with this function's arguments
        if add_cbar:
            legend = True
        else:
            legend = False

        if "legend" in list(kwargs.keys()):
            legend = kwargs.pop("legend")

        # Get colormap arguments that might have been passed in the keyword args
        if "legend_kwds" in list(kwargs.keys()) and legend:
            legend_kwds = kwargs.pop("legend_kwds")
            if cbar_title is not None:
                legend_kwds.update({"label": cbar_title})  # Pad updates depending on figsize during plot,
        else:
            if cbar_title is not None:
                legend_kwds = {"label": cbar_title}
            else:
                legend_kwds = None

        # Add colorbar
        if add_cbar or cbar_title:
            divider = make_axes_locatable(ax0)
            cax = divider.append_axes("right", size="5%", pad="2%")
            norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
            cbar = matplotlib.colorbar.ColorbarBase(
                cax, cmap=cmap, norm=norm
            )  # , orientation="horizontal", ticklocation="top")
            cbar.solids.set_alpha(alpha)
        else:
            cax = None
            cbar = None

        # Plot
        vect_reproj.ds.plot(
            ax=ax0,
            cax=cax,
            column=column,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            alpha=alpha,
            legend=legend,
            legend_kwds=legend_kwds,
            **kwargs,
        )
        plt.sca(ax0)

        # if savefig_fname filled, save the plot
        if savefig_fname:
            plt.savefig(savefig_fname)

        # If returning axes
        if return_axes:
            return ax0, cax
        else:
            return None

    def __add__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:
        """
        Sum two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data + other_data
        return self.copy(new_array=out_data)  # type: ignore

    # Skip Mypy not resolving forward operator typing with NumPy numbers: https://github.com/python/mypy/issues/11595
    def __radd__(self: PointCloud, other: NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Sum two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        For when other is first item in the operation (e.g. 1 + rst).
        """
        return self.__add__(other)  # type: ignore

    def __neg__(self: PointCloud) -> PointCloud:
        """
        Take the point cloud negation.

        Returns a point cloud with -self.data.
        """
        return self.copy(-self.data)

    def __sub__(self, other: PointCloud | NDArrayNum | Number) -> PointCloud:
        """
        Subtract two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data - other_data
        return self.copy(new_array=out_data)  # type: ignore

    # Skip Mypy not resolving forward operator typing with NumPy numbers: https://github.com/python/mypy/issues/11595
    def __rsub__(self: PointCloud, other: NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Subtract two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        For when other is first item in the operation (e.g. 1 - rst).
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = other_data - self.data
        return self.copy(new_array=out_data)  # type: ignore

    def __mul__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:
        """
        Multiply two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data * other_data
        return self.copy(new_array=out_data)  # type: ignore

    # Skip Mypy not resolving forward operator typing with NumPy numbers: https://github.com/python/mypy/issues/11595
    def __rmul__(self: PointCloud, other: NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Multiply two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        For when other is first item in the operation (e.g. 2 * rst).
        """
        return self.__mul__(other)  # type: ignore

    def __truediv__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:
        """
        True division of two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data / other_data
        return self.copy(new_array=out_data)  # type: ignore

    # Skip Mypy not resolving forward operator typing with NumPy numbers: https://github.com/python/mypy/issues/11595
    def __rtruediv__(self: PointCloud, other: NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        True division of two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        For when other is first item in the operation (e.g. 1/rst).
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = other_data / self.data
        return self.copy(new_array=out_data)  # type: ignore

    def __floordiv__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:
        """
        Floor division of two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data // other_data  # type: ignore
        return self.copy(new_array=out_data)

    # Skip Mypy not resolving forward operator typing with NumPy numbers: https://github.com/python/mypy/issues/11595
    def __rfloordiv__(self: PointCloud, other: NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Floor division of two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        For when other is first item in the operation (e.g. 1/rst).
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = other_data // self.data  # type: ignore
        return self.copy(new_array=out_data)

    def __mod__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:
        """
        Modulo of two point clouds, or a point cloud and a numpy array, or a point cloud and single number.

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data % other_data  # type: ignore
        return self.copy(new_array=out_data)

    def __pow__(self: PointCloud, power: int | float) -> PointCloud:
        """
        Power of a point cloud to a number.
        """
        # Check that input is a number
        if not isinstance(power, (float, int, np.floating, np.integer)):
            raise ValueError("Power needs to be a number.")

        # Calculate the product of arrays and save to new point cloud
        out_data = self.data**power
        return self.copy(new_array=out_data)

    def __eq__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Element-wise equality of two point clouds, or a point cloud and a numpy array, or a point cloud and single
        number.

        This operation casts the result into a mask (boolean Raster).

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data == other_data
        return self.copy(new_array=out_data)

    def __ne__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Element-wise negation of two point clouds, or a point cloud and a numpy array, or a point cloud and single
        number.

        This operation casts the result into a mask (boolean Raster).

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data != other_data
        return self.copy(new_array=out_data)

    def __lt__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Element-wise lower than comparison of two point clouds, or a point cloud and a numpy array,
        or a point cloud and single number.

        This operation casts the result into a mask (boolean Raster).

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data < other_data
        return self.copy(new_array=out_data)

    def __le__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Element-wise lower or equal comparison of two point clouds, or a point cloud and a numpy array,
        or a point cloud and single number.

        This operation casts the result into a mask (boolean Raster).

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data <= other_data
        return self.copy(new_array=out_data)

    def __gt__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Element-wise greater than comparison of two point clouds, or a point cloud and a numpy array,
        or a point cloud and single number.

        This operation casts the result into a mask (boolean Raster).

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data > other_data
        return self.copy(new_array=out_data)

    def __ge__(self: PointCloud, other: PointCloud | NDArrayNum | Number) -> PointCloud:  # type: ignore
        """
        Element-wise greater or equal comparison of two point clouds, or a point cloud and a numpy array,
        or a point cloud and single number.

        This operation casts the result into a mask (boolean Raster).

        If other is a point cloud, it must have the same shape, coordinates and crs as self.
        If other is a np.ndarray, it must have the same shape.
        Otherwise, other must be a single number.
        """
        other_data = _cast_numeric_array_pointcloud(self, other, operation_name="an arithmetic operation")
        out_data = self.data >= other_data
        return self.copy(new_array=out_data)

    def __and__(self: PointCloud, other: PointCloud | NDArrayBool) -> PointCloud:
        """Bitwise and between masks, or a mask and an array."""
        other_data = _cast_numeric_array_pointcloud(
            self, other, operation_name="an arithmetic operation"  # type: ignore
        )

        return self.copy(self.data & other_data)  # type: ignore

    def __rand__(self: PointCloud, other: PointCloud | NDArrayBool) -> PointCloud:
        """Bitwise and between masks, or a mask and an array."""

        return self.__and__(other)

    def __or__(self: PointCloud, other: PointCloud | NDArrayBool) -> PointCloud:
        """Bitwise or between masks, or a mask and an array."""

        other_data = _cast_numeric_array_pointcloud(
            self, other, operation_name="an arithmetic operation"  # type: ignore
        )

        return self.copy(self.data | other_data)  # type: ignore

    def __ror__(self: PointCloud, other: PointCloud | NDArrayBool) -> PointCloud:
        """Bitwise or between masks, or a mask and an array."""

        return self.__or__(other)

    def __xor__(self: PointCloud, other: PointCloud | NDArrayBool) -> PointCloud:
        """Bitwise xor between masks, or a mask and an array."""

        other_data = _cast_numeric_array_pointcloud(
            self, other, operation_name="an arithmetic operation"  # type: ignore
        )

        return self.copy(self.data ^ other_data)  # type: ignore

    def __rxor__(self: PointCloud, other: PointCloud | NDArrayBool) -> PointCloud:
        """Bitwise xor between masks, or a mask and an array."""

        return self.__xor__(other)

    def __invert__(self: PointCloud) -> PointCloud:
        """Bitwise inversion of a mask."""

        return self.copy(~self.data)

    def pointcloud_equal(self, other: PointCloud, **kwargs: Any) -> bool:
        """
        Check if two point clouds are equal.

        This means that:
        - The two vectors (geodataframes) are equal.
        - The data column is the same for both point clouds.

        Keyword arguments are passed to geopandas.assert_geodataframe_equal.
        """

        # Vector equality
        vector_eq = super().vector_equal(other, **kwargs)
        # Data column equality
        data_column_eq = self.data_column == other.data_column

        return all([vector_eq, data_column_eq])

    def georeferenced_coords_equal(self: PointCloud, pc: PointCloud) -> bool:
        """
        Check that point cloud X/Y coordinates and CRS are equal.

        :param pc: Another pointcloud.

        :return: Whether the two objects have the same georeferenced points.
        """

        return all(
            [
                self.crs == pc.crs,
                np.array_equal(self.geometry.x.values, pc.geometry.x.values),
                np.array_equal(self.geometry.y.values, pc.geometry.y.values),
            ]
        )

    @overload
    def get_stats(
        self,
        stats_name: str | Callable[[NDArrayNum], np.floating[Any]],
    ) -> np.floating[Any]: ...

    @overload
    def get_stats(
        self,
        stats_name: list[str | Callable[[NDArrayNum], np.floating[Any]]] | None = None,
    ) -> dict[str, np.floating[Any]]: ...

    @profiler.profile("geoutils.pointcloud.pointcloud.get_stats", memprof=True)  # type: ignore
    def get_stats(
        self,
        stats_name: (
            str | Callable[[NDArrayNum], np.floating[Any]] | list[str | Callable[[NDArrayNum], np.floating[Any]]] | None
        ) = None,
    ) -> np.floating[Any] | dict[str, np.floating[Any]]:
        """
        Retrieve specified statistics or all available statistics for the point cloud data. Allows passing custom
        callables to calculate custom stats.

        Common statistics for an N-D array :

        - Mean: arithmetic mean of the data, ignoring masked values.
        - Median: middle value when the valid data points are sorted in increasing order, ignoring masked values.
        - Max: maximum value among the data, ignoring masked values.
        - Min: minimum value among the data, ignoring masked values.
        - Sum: sum of all data, ignoring masked values.
        - Sum of squares: sum of the squares of all data, ignoring masked values.
        - 90th percentile: point below which 90% of the data falls, ignoring masked values.
        - IQR (Interquartile Range): difference between the 75th and 25th percentile of a dataset,
        ignoring masked values.
        - LE90 (Linear Error with 90% confidence): difference between the 95th and 5th percentiles of a dataset,
        representing the range within which 90% of the data points lie. Ignore masked values.
        - NMAD (Normalized Median Absolute Deviation): robust measure of variability in the data, less sensitive to
        outliers compared to standard deviation. Ignore masked values.
        - RMSE (Root Mean Square Error): commonly used to express the magnitude of errors or variability and can give
        insight into the spread of the data. Only relevant when the raster represents a difference of two objects.
        Ignore masked values.
        - Std (Standard deviation): measures the spread or dispersion of the data around the mean,
        ignoring masked values.
        - Valid count: number of finite data points in the array. It counts the non-masked elements.
        - Total count: total size of the raster.
        - Percentage valid points: ratio between Valid count and Total count.

        For all statistics up to "Std", functions from numpy.ma module are used (directly or in the calculation) in case
        of a masked array, numpy module otherwise.

        "Valid count" represents all non zero and not masked pixels in the input data (final_count_nonzero).
        Numpy Masked functions is used is this case or if the Point Cloud was already a masked array.
        Percentage valid points is calculated accordingly.

        If an inlier mask is passed:
        - Total inlier count: number of data points in the inlier mask.
        - Valid inlier count: number of unmasked data points in the array after applying the inlier mask.
        - Percentage inlier points: ratio between Valid inlier count and Valid count. Useful for classification
        statistics.
        - Percentage valid inlier points: ratio between Valid inlier count and Total inlier count.

        They are all computed based on the previously stated final_count_nonzero.

        Callable functions are supported as well.

        :param stats_name: Name or list of names of the statistics to retrieve. If None, all statistics are returned.
            Accepted names include:
            `mean`, `median`, `max`, `min`, `sum`, `sum of squares`, `90th percentile`, `LE90`, `nmad`, `rmse`,
            `std`, `valid count`, `total count`, `percentage valid points` and if an inlier mask is passed :
            `valid inlier count`, `total inlier count`, `percentage inlier point`, `percentage valid inlier points`.
            Custom callables can also be provided.
        :returns: The requested statistic or a dictionary of statistics if multiple or all are requested.
        """

        # Force load if not loaded
        if not self.is_loaded:
            self.load()

        data = self.data

        # Given list or all attributes to compute if None
        if isinstance(stats_name, list) or stats_name is None:
            return _statistics(data, stats_name)  # type: ignore
        else:
            # Single attribute to compute
            if isinstance(stats_name, str):
                return _statistics(data, [stats_name])[stats_name]  # type: ignore
            elif callable(stats_name):
                return stats_name(data)  # type: ignore
            else:
                logging.warning("Statistic name '%s' is a not recognized string", stats_name)

    @overload
    def subsample(
        self,
        subsample: int | float,
        return_indices: Literal[False] = False,
        *,
        random_state: int | np.random.Generator | None = None,
    ) -> NDArrayNum: ...

    @overload
    def subsample(
        self,
        subsample: int | float,
        return_indices: Literal[True],
        *,
        random_state: int | np.random.Generator | None = None,
    ) -> tuple[NDArrayNum, ...]: ...

    @overload
    def subsample(
        self,
        subsample: float | int,
        return_indices: bool = False,
        random_state: int | np.random.Generator | None = None,
    ) -> NDArrayNum | tuple[NDArrayNum, ...]: ...

    @profiler.profile("geoutils.pointcloud.pointcloud.subsample", memprof=True)  # type: ignore
    def subsample(
        self,
        subsample: float | int,
        return_indices: bool = False,
        random_state: int | np.random.Generator | None = None,
    ) -> NDArrayNum | tuple[NDArrayNum, ...]:
        """
        Randomly sample the point cloud. Only valid values are considered.

        :param subsample: Subsample size. If <= 1, a fraction of the total pixels to extract.
            If > 1, the number of pixels.
        :param return_indices: Whether to return the extracted indices only.
        :param random_state: Random state or seed number.

        :return: Array of sampled valid values, or array of sampled indices.
        """

        return subsample_array(
            array=self.data, subsample=subsample, return_indices=return_indices, random_state=random_state
        )

    @profiler.profile("geoutils.pointcloud.pointcloud.grid", memprof=True)  # type: ignore
    def grid(
        self,
        ref: gu.Raster | None = None,
        grid_coords: tuple[NDArrayNum, NDArrayNum] | None = None,
        res: float | tuple[float, float] | None = None,
        resampling: Literal["nearest", "linear", "cubic"] = "linear",
        dist_nodata_pixel: float = 1.0,
        nodata: int | float = -9999,
    ) -> gu.Raster:
        """
        Grid point cloud into a raster.

        Output grid can be defined either by passing a reference raster to match, or by passing output grid coordinates
        for X/Y (must be regular in each dimension), or by specifying an output grid resolution (which uses the
        upper-left bounds of point cloud to define the start of the grid).

        :param ref: Reference raster to match (if output grid coordinates or output resolution undefined).
        :param grid_coords: Output grid coordinates in X and Y (if reference raster or output resolution undefined).
        :param res: Output resolution (if reference raster or output grid coordinates undefined).
        :param resampling: Resampling method within delauney triangles (defaults to linear).
        :param dist_nodata_pixel: Distance from the point cloud after which grid cells are filled by nodata values,
            expressed in number of pixels.
        :param nodata: Nodata value of output raster (defaults to -9999).

        :return: Raster from gridded point cloud.
        """

        if isinstance(ref, gu.Raster):
            if grid_coords is not None:
                warnings.warn(
                    "Both reference point cloud and grid coordinates were passed for gridding, "
                    "using only the reference point cloud."
                )
            grid_coords = ref.coords(grid=False)
        else:
            if res is not None:
                xsize = (self.bounds.right - self.bounds.left) / res
                ysize = (self.bounds.top - self.bounds.bottom) / res
                transform = from_origin(west=self.bounds.left, north=self.bounds.top, xsize=xsize, ysize=ysize)
                grid_coords = _coords(transform=transform, shape=(ysize, xsize), grid=False, area_or_point=None)
            else:
                grid_coords = grid_coords

        array, transform = _grid_pointcloud(
            self.ds,
            grid_coords=grid_coords,
            data_column_name=self.data_column,
            resampling=resampling,
            dist_nodata_pixel=dist_nodata_pixel,
        )

        return gu.Raster.from_array(data=array, transform=transform, crs=self.crs, nodata=nodata)



"""
Test tools involving multiple rasters.
"""

from __future__ import annotations

import warnings
from typing import Any, Callable

import numpy as np
import pyproj
import pytest
import rasterio as rio
from pytest_lazy_fixtures import lf as lazy_fixtures

import geoutils as gu
from geoutils import examples
from geoutils._typing import NDArrayNum
from geoutils.raster import RasterType
from geoutils.raster.raster import _default_nodata


class RealImageStack:
    """
    Real test cases for stacking and merging images
    Split an image with some overlap, then stack/merge it, and validate bounds and shape.
    Param `cls` is used to set the type of the output, e.g. gu.Raster (default).
    """

    def __init__(
        self, image: str, cls: Callable[[str], RasterType] = gu.Raster, different_crs: pyproj.CRS | None = None
    ) -> None:

        warnings.filterwarnings("ignore", category=UserWarning, message="For reprojection, nodata must be set.*")

        img = cls(examples.get_path_test(image))
        self.img = img

        # Find the easting midpoint of the img
        x_midpoint = np.mean([img.bounds.right, img.bounds.left])
        x_midpoint -= (x_midpoint - img.bounds.left) % img.res[0]

        # Cut the img into two imgs that slightly overlap each other.
        self.img1 = img.copy()
        self.img1.crop(
            rio.coords.BoundingBox(
                right=x_midpoint + img.res[0] * 3, left=img.bounds.left, top=img.bounds.top, bottom=img.bounds.bottom
            ),
            inplace=True,
        )
        self.img2 = img.copy()
        self.img2.crop(
            rio.coords.BoundingBox(
                left=x_midpoint - img.res[0] * 3, right=img.bounds.right, top=img.bounds.top, bottom=img.bounds.bottom
            ),
            inplace=True,
        )
        if different_crs:
            self.img2 = self.img2.reproject(crs=different_crs, resampling="nearest")

        # To check that use_ref_bounds work - create a img that do not cover the whole extent
        self.img3 = img.copy()
        self.img3.crop(
            rio.coords.BoundingBox(
                left=x_midpoint - img.res[0] * 3,
                right=img.bounds.right - img.res[0] * 2,
                top=img.bounds.top,
                bottom=img.bounds.bottom,
            ),
            inplace=True,
        )


class SyntheticImageStack:
    """
    Synthetic image stack for tests

    Create a small synthetic example, where one can specify nodata value, values in second image (and potentially more
    in the future).
    """

    def __init__(self, nodata: int | float, img2_value: int | float):

        shape = (10, 10)
        data_int = np.ones(shape).astype(np.uint16)
        data_mask = np.zeros(shape).astype(bool)
        data_masked = np.ma.masked_array(data=data_int, mask=data_mask, fill_value=nodata)
        img = gu.Raster.from_array(
            data=data_masked,
            transform=rio.transform.Affine(
                1000.0,
                0.0,
                1_000_000.0,
                0.0,
                -1000.0,
                1_000_000.0,
            ),
            crs=pyproj.CRS.from_string("EPSG:3857"),
            nodata=nodata,
        )
        self.img = img

        # Find the easting midpoint of the img
        x_midpoint = np.mean([self.img.bounds.right, self.img.bounds.left])
        x_midpoint -= (x_midpoint - self.img.bounds.left) % self.img.res[0]

        # Cut the img into two imgs that slightly overlap each other.
        self.img1 = img.copy()
        self.img1.crop(
            rio.coords.BoundingBox(
                right=x_midpoint + img.res[0] * 3, left=img.bounds.left, top=img.bounds.top, bottom=img.bounds.bottom
            ),
            inplace=True,
        )
        self.img2 = img.copy()
        self.img2.crop(
            rio.coords.BoundingBox(
                left=x_midpoint - img.res[0] * 3, right=img.bounds.right, top=img.bounds.top, bottom=img.bounds.bottom
            ),
            inplace=True,
        )

        # Define a second raster with only 5s and the value defined above
        self.img2[:5, :5] = img2_value

        self.img3 = self.img1.copy()
        self.img3.crop(
            rio.coords.BoundingBox(
                left=x_midpoint - self.img.res[0] * 3,
                right=self.img.bounds.right - self.img.res[0] * 2,
                top=self.img.bounds.top,
                bottom=self.img.bounds.bottom,
            ),
            inplace=True,
        )


@pytest.fixture
def images_1d():  # type: ignore
    return RealImageStack("everest_landsat_b4")


@pytest.fixture
def images_different_crs():  # type: ignore
    return RealImageStack("everest_landsat_b4", different_crs=4326)


@pytest.fixture
def images_3d():  # type: ignore
    return RealImageStack("everest_landsat_rgb")


@pytest.fixture
def images_nodata_zero():  # type: ignore
    return SyntheticImageStack(nodata=0, img2_value=65534)


class TestMultiRaster:
    @pytest.mark.parametrize(
        "rasters",
        [
            lazy_fixtures("images_1d"),
            lazy_fixtures("images_different_crs"),
            lazy_fixtures("images_3d"),
            lazy_fixtures("images_nodata_zero"),
        ],
    )  # type: ignore
    def test_stack_rasters(self, rasters: Any) -> None:  # type: ignore
        """Test stack_rasters"""

        # Silence the reprojection warning for default nodata value
        warnings.filterwarnings(
            "ignore", category=UserWarning, message="New nodata value cells already exist in the data array.*"
        )
        warnings.filterwarnings("ignore", category=UserWarning, message="For reprojection, nodata must be set.*")
        warnings.filterwarnings("ignore", category=UserWarning, message="Unmasked values equal to*")

        # Merge the two overlapping DEMs and check that output bounds and shape is correct
        if rasters.img1.count > 1:
            # Check warning is raised once
            with pytest.warns(
                expected_warning=UserWarning,
                match="Some input Rasters have multiple bands, only their first band will be used.",
            ):
                stacked_img = gu.raster.stack_rasters([rasters.img1, rasters.img2])
            # Then ignore the other ones
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message="Some input Rasters have multiple bands, only their first band will be used.",
            )

        else:
            stacked_img = gu.raster.stack_rasters([rasters.img1, rasters.img2])

        assert stacked_img.count == 2
        # If the rasters were in a different projection, the final shape can vary by 1 pixel
        if not all(rast.crs == rasters.img.crs for rast in [rasters.img1, rasters.img2]):
            assert rasters.img.height == pytest.approx(stacked_img.height, abs=1)
            assert rasters.img.width == pytest.approx(stacked_img.width, abs=1)
        else:
            assert rasters.img.shape == stacked_img.shape
        assert isinstance(stacked_img, gu.Raster)  # Check output object is always Raster, whatever input was given
        assert np.count_nonzero(np.isnan(stacked_img.data)) == 0  # Check no NaNs introduced

        merged_bounds = gu.projtools.merge_bounds(
            [rasters.img1.bounds, rasters.img2.get_bounds_projected(rasters.img1.crs)], resolution=rasters.img1.res[0]
        )
        assert merged_bounds == stacked_img.bounds

        nodata_ref = rasters.img1.nodata
        # Check that reference works with input Raster
        stacked_img = gu.raster.stack_rasters([rasters.img1, rasters.img2], reference=rasters.img, use_ref_bounds=True)
        assert rasters.img.bounds == stacked_img.bounds

        # Others than int or gu.Raster should raise a ValueError
        try:
            stacked_img = gu.raster.stack_rasters([rasters.img1, rasters.img2], reference="a string")
        except ValueError as exception:
            if "reference should be" not in str(exception):
                raise exception

        # Check that use_ref_bounds works - use a img that do not cover the whole extent

        # This case should not preserve original extent
        stacked_img = gu.raster.stack_rasters([rasters.img1, rasters.img3])
        assert stacked_img.bounds != rasters.img.bounds

        # This case should preserve original extent
        stacked_img2 = gu.raster.stack_rasters([rasters.img1, rasters.img3], reference=rasters.img, use_ref_bounds=True)
        assert stacked_img2.bounds == rasters.img.bounds

        # This case should preserve unique data values through "nearest" resampling
        rasters.img1[:] = 5
        rasters.img1[0:5, 0:5] = 1
        rasters.img2 = rasters.img1.translate(0.5, 0.5, distance_unit="pixel")
        stacked_img = gu.raster.stack_rasters([rasters.img1, rasters.img2], resampling_method="nearest")
        assert np.array_equal(np.unique(stacked_img.data.compressed()), np.array([1, 5]))
        # But not this case with a shifted raster resampled with "bilinear"
        stacked_img = gu.raster.stack_rasters([rasters.img1, rasters.img2], resampling_method="bilinear")
        assert not np.array_equal(np.unique(stacked_img.data.compressed()), np.array([1, 5]))

        # Check input nodata is not modified inplace (issue 609)
        new_nodata_ref = rasters.img1.nodata
        assert nodata_ref == new_nodata_ref

        # Check nodata value output is consistent with reference input
        if nodata_ref is not None:
            assert stacked_img.nodata == nodata_ref
        else:
            assert stacked_img.nodata == _default_nodata(rasters.img1.dtype)

    @pytest.mark.parametrize(
        "rasters",
        [
            lazy_fixtures("images_1d"),
            lazy_fixtures("images_3d"),
            lazy_fixtures("images_different_crs"),
        ],
    )  # type: ignore
    def test_merge_rasters(self, rasters: Any) -> None:  # type: ignore
        """Test merge_rasters"""
        # Merge the two overlapping DEMs and check that it closely resembles the initial DEM

        # Silence the reprojection warning for default nodata value
        warnings.filterwarnings(
            "ignore", category=UserWarning, message="New nodata value cells already exist in the data array.*"
        )
        warnings.filterwarnings("ignore", category=UserWarning, message="For reprojection, nodata must be set.*")
        warnings.filterwarnings("ignore", category=UserWarning, message="Unmasked values equal to*")

        # Ignore warning already checked in test_stack_rasters
        if rasters.img1.count > 1:
            warnings.filterwarnings(
                "ignore",
                category=UserWarning,
                message="Some input Rasters have multiple bands, only their first band will be used.",
            )

        merged_img = gu.raster.merge_rasters([rasters.img1, rasters.img2], merge_algorithm=np.nanmean)

        if not all(rast.crs == rasters.img.crs for rast in [rasters.img1, rasters.img2]):
            assert rasters.img.height == pytest.approx(merged_img.height, abs=1)
            assert rasters.img.width == pytest.approx(merged_img.width, abs=1)
        else:
            assert rasters.img.shape == merged_img.shape

        merged_bounds = gu.projtools.merge_bounds(
            [rasters.img1.bounds, rasters.img2.get_bounds_projected(rasters.img1.crs)], resolution=rasters.img1.res[0]
        )
        assert merged_bounds == merged_img.bounds

        assert np.count_nonzero(np.isnan(merged_img.data)) == 0  # Check no NaNs introduced

        # Check that reference works
        merged_img2 = gu.raster.merge_rasters([rasters.img1, rasters.img2], reference=rasters.img, use_ref_bounds=True)
        # Check that only works if CRS were the same
        if all(rast.crs == rasters.img.crs for rast in [rasters.img1, rasters.img2]):
            assert merged_img2 == merged_img

        # For merge algo: function not supporting the axis keyword argument but raising the right "axis" type error
        def custom_func(x: NDArrayNum) -> NDArrayNum:
            return np.logical_and(*x)

        gu.raster.merge_rasters([rasters.img1, rasters.img2], merge_algorithm=custom_func)

    @pytest.mark.parametrize(
        "rasters",
        [
            lazy_fixtures("images_1d"),
            lazy_fixtures("images_3d"),
        ],
    )  # type: ignore
    def test_merge_rasters__errors(self, rasters: Any) -> None:
        """Test errors of merge raster are properly raised."""

        # For merge algo: function that raises another type error than the expect axis error
        msg = "not the right axis message"

        def custom_func(x: NDArrayNum) -> None:
            raise TypeError(msg)

        with pytest.raises(TypeError, match=msg):
            gu.raster.merge_rasters([rasters.img1, rasters.img2], merge_algorithm=custom_func)

    # Group rasters for for testing `load_multiple_rasters`
    # two overlapping, single band rasters
    # two overlapping, 1 and 3 band rasters
    # three overlapping rasters
    # TODO: add a case with different CRS - issue raised #310
    raster_groups = [
        [gu.examples.get_path_test("everest_landsat_b4"), gu.examples.get_path_test("everest_landsat_b4_cropped")],
        [gu.examples.get_path_test("everest_landsat_rgb"), gu.examples.get_path_test("everest_landsat_b4_cropped")],
        [
            gu.examples.get_path_test("everest_landsat_b4"),
            gu.examples.get_path_test("everest_landsat_rgb"),
            gu.examples.get_path_test("everest_landsat_b4_cropped"),
        ],
    ]

    @pytest.mark.parametrize("raster_paths", raster_groups)  # type: ignore
    def test_load_multiple_overlap(self, raster_paths: list[str]) -> None:
        """
        Test load_multiple_rasters functionalities, when rasters overlap -> no warning is raised
        """
        # - Test that with crop=False and ref_grid=None, rasters are simply loaded - #
        output_rst: list[gu.Raster] = gu.raster.load_multiple_rasters(raster_paths, crop=False, ref_grid=None)
        for k, rst in enumerate(output_rst):
            assert rst.is_loaded
            rst2 = gu.Raster(raster_paths[k])
            assert rst.raster_equal(rst2)

        # - Test that with crop=True and ref_grid=None, rasters are cropped only in area of overlap - #
        output_rst = gu.raster.load_multiple_rasters(raster_paths, crop=True, ref_grid=None)
        ref_crs = gu.Raster(raster_paths[0], load_data=False).crs

        # Save original and new bounds (as polygons) in the reference CRS
        orig_poly_bounds = []
        new_poly_bounds = []
        for k, rst in enumerate(output_rst):
            assert rst.is_loaded
            rst2 = gu.Raster(raster_paths[k], load_data=False)
            assert rst.crs == rst2.crs  # CRS should not have been changed
            orig_poly_bounds.append(gu.projtools.bounds2poly(rst2.bounds, rst.crs, ref_crs))
            new_poly_bounds.append(gu.projtools.bounds2poly(rst.bounds, rst.crs, ref_crs))

        # Check that the new bounds are contained in the original bounds of all rasters
        for poly1 in new_poly_bounds:
            for poly2 in orig_poly_bounds:
                assert poly2.contains(poly1)

        # - Test that with crop=False and ref_grid=0, rasters all have the same grid as the reference - #
        # For the landsat test case, a warning will be raised because nodata is None
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            output_rst = gu.raster.load_multiple_rasters(raster_paths, crop=False, ref_grid=0)

        ref_rst = gu.Raster(raster_paths[0], load_data=False)
        for k, rst in enumerate(output_rst):
            rst2 = gu.Raster(raster_paths[k], load_data=False)
            assert rst.is_loaded
            # Georeferences are forced to ref
            assert rst.crs == ref_rst.crs
            assert rst.shape == ref_rst.shape
            assert rst.transform == ref_rst.transform
            # Original number of bounds and nodata must be preserved
            assert rst.count == rst2.count
            if rst2.nodata is not None:  # if none, reprojection with set a default value
                assert rst.nodata == rst2.nodata

    raster_groups = [
        [gu.examples.get_path_test("everest_landsat_b4"), gu.examples.get_path_test("exploradores_aster_dem")],
    ]

    @pytest.mark.parametrize("raster_paths", raster_groups)  # type: ignore
    def test_load_multiple_no_overlap(self, raster_paths: list[str]) -> None:
        """
        Test load_multiple_rasters functionalities with rasters that do not overlap -> raises warning in certain cases
        """
        # - With crop=False and ref_grid=None, rasters are simply loaded - #
        output_rst: list[gu.Raster] = gu.raster.load_multiple_rasters(raster_paths, crop=False, ref_grid=None)
        for k, rst in enumerate(output_rst):
            assert rst.is_loaded
            rst2 = gu.Raster(raster_paths[k])
            assert rst.raster_equal(rst2)

        # - With crop=True -> should raise a warning - #
        with pytest.warns(UserWarning, match="Intersection is void, returning unloaded rasters."):
            output_rst = gu.raster.load_multiple_rasters(raster_paths, crop=True, ref_grid=None)

        # - Should work with crop=False and ref_grid=0 - #
        # For the landsat test case, a warning will be raised because nodata is None
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)
            output_rst = gu.raster.load_multiple_rasters(raster_paths, crop=False, ref_grid=0)

        ref_rst = gu.Raster(raster_paths[0], load_data=False)
        for k, rst in enumerate(output_rst):
            rst2 = gu.Raster(raster_paths[k], load_data=False)
            assert rst.is_loaded
            # Georeferences are forced to ref
            assert rst.crs == ref_rst.crs
            assert rst.shape == ref_rst.shape
            assert rst.transform == ref_rst.transform
            # Original number of bounds and nodata must be preserved
            assert rst.count == rst2.count
            if rst2.nodata is not None:  # if none, reprojection with set a default value
                assert rst.nodata == rst2.nodata
            # Logically, the non overlapping raster should have only masked values
            if k != 0:
                assert np.count_nonzero(~rst.data.mask) == 0



# Copyright (c) 2025 Centre National d'Etudes Spatiales (CNES)
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the cluster configurations."""

import multiprocessing
from multiprocessing.pool import Pool
from typing import Any, Callable, Dict, List, Optional


class ClusterGenerator:
    def __new__(cls, name: str, nb_workers: int = 2) -> "AbstractCluster":  # type: ignore
        """
        Factory method to create different types of clusters based on the name argument.
        - If 'basic' is provided, a BasicCluster is instantiated.
        - Otherwise, an MpCluster (multiprocessing) is created with the given number of workers.
        """
        if name == "basic":
            cluster: AbstractCluster = BasicCluster()
        else:
            cluster = MpCluster(conf={"nb_workers": nb_workers})
        return cluster


class AbstractCluster:
    def __init__(self) -> None:
        """
        Base class for clusters. Initializes the pool attribute.
        Meant to be subclassed and extended.
        """
        self.pool: Optional[Pool] = None

    def __enter__(self) -> "AbstractCluster":
        """Context manager entry point, returning the cluster instance."""
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        """Context manager exit point, ensures the cluster is properly closed."""
        self.close()

    def close(self) -> None:
        """Method to clean up resources. To be implemented by subclasses."""
        raise NotImplementedError("This method should be implemented by subclasses.")

    def launch_task(
        self, fun: Callable[..., Any], args: Optional[List[Any]] = None, kwargs: Optional[Dict[str, Any]] = None
    ) -> Any:
        """
        Method to launch a task. This should be implemented by subclasses.
        :param fun: The function to run.
        :param args: Positional arguments for the function.
        :param kwargs: Keyword arguments for the function.
        """
        raise NotImplementedError("This method should be implemented by subclasses.")

    def get_res(self, future: Any) -> Any:
        """
        Retrieve the result from a launched task. Meant to be subclassed.
        :param future: The future object representing the result of an asynchronous task.
        """
        return future

    def return_wrapper(self) -> None:
        """Wrapper for returned values, should be customized in subclasses if needed."""
        raise NotImplementedError("This method should be implemented by subclasses.")

    def tile_retriever(self, res: Any) -> None:
        """Method to retrieve and process tiles, specific to subclasses."""
        raise NotImplementedError("This method should be implemented by subclasses.")


class BasicCluster(AbstractCluster):
    def launch_task(
        self, fun: Callable[..., Any], args: Optional[List[Any]] = None, kwargs: Optional[Dict[str, Any]] = None
    ) -> Any:
        """
        Launches a task synchronously in a basic cluster (no multiprocessing).
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        return fun(*args, **kwargs)


class MpCluster(AbstractCluster):
    def __init__(self, conf: Optional[Dict[str, Any]] = None) -> None:
        """
        Initializes a multiprocessing cluster.
        :param conf: Configuration dictionary, which may contain the number of workers.
        """
        super().__init__()
        nb_workers = 1
        if conf is not None:
            nb_workers = conf.get("nb_workers", 1)
        # Using the 'forkserver' context for more controlled process handling
        ctx_in_main = multiprocessing.get_context("fork")
        # Create a pool of workers with max 10 tasks per child process
        self.pool = ctx_in_main.Pool(processes=nb_workers, maxtasksperchild=10)

    def close(self) -> None:
        """Closes the multiprocessing pool by terminating and joining workers."""
        if self.pool is not None:
            self.pool.terminate()
            self.pool.join()

    def launch_task(
        self, fun: Callable[..., Any], args: Optional[List[Any]] = None, kwargs: Optional[Dict[str, Any]] = None
    ) -> Any:
        """
        Launches a task asynchronously in the multiprocessing pool.
        :param fun: The function to execute in parallel.
        :param args: The positional arguments for the function.
        :param kwargs: The keyword arguments for the function.

        :return: an asynchronous result (future object).
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        if self.pool is not None:
            return self.pool.apply_async(fun, args=args, kwds=kwargs)

    def get_res(self, future: Any) -> Any:
        """
        Retrieves the result of a completed asynchronous task.
        """
        return future.get(timeout=5000)



"""
Test the statistical estimator module.
"""

import numpy as np
import pytest
import scipy

from geoutils import Raster, examples
from geoutils.stats import linear_error, nmad, rmse, sum_square


class TestEstimators:
    landsat_b4_path = examples.get_path_test("everest_landsat_b4")
    landsat_raster = Raster(landsat_b4_path)

    def test_nmad(self) -> None:
        """Test NMAD functionality runs on any type of input"""

        # Check that the NMAD is computed the same with a masked array or NaN array, and is equal to scipy nmad
        nmad_ma = nmad(self.landsat_raster.data)
        nmad_array = nmad(self.landsat_raster.get_nanarray(floating_dtype="float64"))
        nmad_scipy = scipy.stats.median_abs_deviation(
            self.landsat_raster.get_nanarray(floating_dtype="float64"), axis=None, scale="normal"
        )

        assert nmad_ma == nmad_array
        assert nmad_ma.round(2) == nmad_scipy.round(2)

        # Check that the scaling factor works
        nmad_1 = nmad(self.landsat_raster.data, nfact=1)
        nmad_2 = nmad(self.landsat_raster.data, nfact=2)

        assert nmad_1 * 2 == nmad_2

    def test_linear_error(self) -> None:
        """Test linear error (LE) functionality runs on any type of input"""

        # Compute LE on the landsat raster data for a default interval (LE90)
        le_ma = linear_error(self.landsat_raster.data)
        le_nan_array = linear_error(self.landsat_raster.get_nanarray())

        # Assert the LE90 is computed the same for masked array and NaN array
        assert le_ma == le_nan_array

        # Check that the function works for different intervals
        le90 = linear_error(self.landsat_raster.data, interval=90)
        le50 = linear_error(self.landsat_raster.data, interval=50)

        # Verify that LE50 (interquartile range) is smaller than LE90
        assert le50 < le90

        # Test a known dataset
        test_data = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        le90_test = linear_error(test_data, interval=90)
        le50_test = linear_error(test_data, interval=50)

        assert le90_test == 9
        assert le50_test == 5

        # Test masked arrays with invalid data (should ignore NaNs/masked values)
        masked_data = np.ma.masked_array(test_data, mask=[0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        le90_masked = linear_error(masked_data, interval=90)
        le50_masked = linear_error(masked_data, interval=50)

        assert le90_masked == pytest.approx(4.5)
        assert le50_masked == 2.5

    def test_rmse(self) -> None:
        """Test RMSE functionality runs on any type of input"""

        test_data = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        # Test masked arrays with invalid data (should ignore NaNs/masked values)
        masked_data = np.ma.masked_array(test_data, mask=[0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        # Corresponding array
        test_data_crop = np.array([0, 1, 2, 3, 4, 5])

        rmse_data = rmse(test_data_crop)
        rmse_masked_data = rmse(masked_data)

        assert rmse_data == rmse_masked_data
        assert rmse_data == pytest.approx(3.0276503540974917)

    def test_sum_square(self) -> None:
        """Test Sum Square functionality runs on any type of input"""

        test_data = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        # Test masked arrays with invalid data (should ignore NaNs/masked values)
        masked_data = np.ma.masked_array(test_data, mask=[0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1])
        # Corresponding array
        test_data_crop = np.array([0, 1, 2, 3, 4, 5])

        sum_square_data = sum_square(test_data_crop)
        sum_square_masked_data = sum_square(masked_data)

        assert sum_square_data == sum_square_masked_data
        assert sum_square_data == 55



"""
Gridding points to raster
=========================

This example demonstrates the gridding of a point cloud into a raster using :func:`~geoutils.PointCloud.gridding`.
"""

# %%
# We open an example point cloud, an elevation dataset in New Zealand.

# sphinx_gallery_thumbnail_number = 2
import geoutils as gu

filename_pc = gu.examples.get_path("coromandel_lidar")
pc = gu.PointCloud(filename_pc, data_column="Z")

# Plot the point cloud
pc.plot(cmap="terrain", cbar_title="Elevation (m)")

# %%
# We generate grid coordinates to interpolate to, alternatively we could pass a raster to use as reference.

import numpy as np

grid_coords = (np.linspace(pc.bounds.left, pc.bounds.right, 100), np.linspace(pc.bounds.bottom, pc.bounds.top, 100))

# %%
# We then perform the interpolation
rast = pc.grid(grid_coords=grid_coords)

# %%
# Finally, we plot the resulting raster

rast.plot(ax="new", cmap="terrain", cbar_title="Elevation (m)")



####
# Part not shown in the doc, to get data files ready
import geoutils

landsat_b4_path = geoutils.examples.get_path("everest_landsat_b4")
landsat_b4_crop_path = geoutils.examples.get_path("everest_landsat_b4_cropped")
geoutils.Raster(landsat_b4_path).to_file("myraster1.tif")
geoutils.Raster(landsat_b4_crop_path).to_file("myraster2.tif")
import warnings

warnings.filterwarnings("ignore", category=UserWarning, message="For reprojection, nodata must be set.*")
warnings.filterwarnings("ignore", category=UserWarning, message="No nodata set*")
warnings.filterwarnings("ignore", category=UserWarning, message="One raster has a pixel interpretation*")
####

import geoutils as gu

# Opening of two rasters
rast1 = gu.Raster("myraster1.tif")
rast2 = gu.Raster("myraster2.tif")

# Reproject 1 to match 2
# (raster 2 not loaded, only metadata)
rast1_reproj = rast1.reproject(ref=rast2)

# Array interfacing and implicit loading
# (raster 2 loads implicitly)
rast_result = (1 + rast2) / rast1_reproj

# Saving
rast_result.to_file("myresult.tif")

####
# Part not shown in the docs, to clean up
import os

for file in ["myraster1.tif", "myraster2.tif", "myresult.tif"]:
    os.remove(file)
####



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Miscellaneous functions for maintenance, documentation and testing."""

from __future__ import annotations

import copy
import functools
import logging
import warnings
from contextlib import contextmanager
from typing import Any, Callable

from packaging.version import Version

import geoutils


@contextmanager  # type: ignore
def silence_rasterio_message(param_name: str, warn_code: str = "CPLE_NotSupported") -> Any:
    """
    Helper to silence logging warnings for parameters that were already supported (such as XSCALE/YSCALE) but not
    yet exposed officially through the API.

    :param param_name: Name of Rasterio parameter.
    :param warn_code: Warning start code within rasterio._env.
    """
    logger = logging.getLogger("rasterio._env")

    class _Filter(logging.Filter):
        def filter(self, record: Any) -> bool:
            return not (
                record.name == "rasterio._env"
                and warn_code in record.getMessage()
                and param_name in record.getMessage()
            )

    flt = _Filter()
    logger.addFilter(flt)
    try:
        yield
    finally:
        logger.removeFilter(flt)


def import_optional(import_name: str, package_name: str | None = None, extra_name: str = "opt") -> Any:
    """
    Helper function to consistently import and raise errors for an optional dependency.

    :param import_name: Name of the dependency to import.
    :param package_name: Name of the package to install (optional, only if different from import name e.g.
        "pyyaml" package is imported as "import yaml".).
    :param extra_name: Name of the extra tag to install the optional dependency from GeoUtils.
    """

    if package_name is None:
        package_name = import_name

    try:
        return __import__(import_name)
    except ImportError as e:
        raise ImportError(
            f"Optional dependency '{package_name}' required. "
            f"Install it directly or through: pip install geoutils[{extra_name}]."
        ) from e


def deprecate(removal_version: Version | None = None, details: str | None = None):  # type: ignore
    """
    Trigger a DeprecationWarning for the decorated function.

    :param func: The function to be deprecated.
    :param removal_version: Optional. The version at which this will be removed.
                            If this version is reached, a ValueError is raised.
    :param details: Optional. A description for why the function was deprecated.

    :triggers DeprecationWarning: For any call to the function.

    :raises ValueError: If 'removal_version' was given and the current version is equal or higher.

    :returns: The decorator to decorate the function.
    """

    def deprecator_func(func):  # type: ignore
        @functools.wraps(func)
        def new_func(*args, **kwargs):  # type: ignore

            # Get current base version (without dev changes)
            current_version = Version(Version(geoutils.__version__).base_version)

            # True if it should warn, False if it should raise an error
            should_warn = removal_version is None or removal_version > current_version

            # Add text depending on the given arguments and 'should_warn'.
            text = (
                f"Call to deprecated function '{func.__name__}'."
                if should_warn
                else f"Deprecated function '{func.__name__}' was removed in {removal_version}."
            )

            # Add the details explanation if it was given, and make sure the sentence is ended.
            if details is not None:
                details_frm = details.strip()
                if details_frm[0].islower():
                    details_frm = details_frm[0].upper() + details_frm[1:]

                text += " " + details_frm

                if not any(text.endswith(c) for c in ".!?"):
                    text += "."

            if should_warn and removal_version is not None:
                text += f" This functionality will be removed in version {removal_version}."
            elif not should_warn:
                text += f" Current version: {geoutils.__version__}."

            if should_warn:
                warnings.warn(text, category=DeprecationWarning, stacklevel=2)
            else:
                raise ValueError(text)

            return func(*args, **kwargs)

        return new_func

    return deprecator_func


def copy_doc(
    old_class: object,
    new_class_name: str,
    origin_class: object | None = None,
    replace_return_series_statement: bool = False,
) -> Callable:  # type: ignore
    """
    A decorator to copy docstring from a class to another class while replacing the docstring.
    Use for parsing GeoPandas' documentation to geoutils.Vector.
    """

    def decorator(decorated: Callable) -> Callable:  # type: ignore

        # Get name of decorated object
        # If object is a property, get name through fget
        try:
            decorated_name = decorated.fget.__name__
        # Otherwise, directly with the name attribute
        except AttributeError:
            decorated_name = decorated.__name__

        # Get parent doc, only if it exists
        # ! VERY IMPORTANT ! The decorator is compiled during package import, if it fails (because a function is
        # deprecated by an upstream package), the whole package breaks on import instead of just this function
        try:
            old_class_doc = getattr(old_class, decorated_name).__doc__
        except AttributeError:
            decorated.__doc__ = "This function documentation does not exist in GeoPandas (likely deprecated)."
            return decorated

        # Remove examples if there are any
        doc_descript = old_class_doc.split("\n\n")[0]

        # Remove duplicate white spaces while keeping lines
        doc_descript = "\n".join([" ".join(spl.split()) for spl in doc_descript.splitlines()])
        # Replace old class output name by the new class output name
        replaced_descript = doc_descript.replace(old_class.__name__, new_class_name)

        # Replace "Return a Series" statement by "Append a Series to Vector" if it exists
        if replace_return_series_statement:
            if replaced_descript[0:20] == "Returns a ``Series``":
                replaced_descript = replaced_descript.replace(
                    "Returns a ``Series``", "Returns or appends to ``Vector`` a ``Series``"
                )
            else:
                replaced_descript += " Can be appended to ``Vector``."

        # Add an intersphinx link to the doc of the previous class
        if origin_class is None:
            orig_class = old_class
        else:
            orig_class = origin_class
        # Get module and old class names
        orig_module_name = orig_class.__module__.split(".")[0]
        old_class_name = orig_class.__name__
        add_link_to_old_class = (
            "\n\nSee more details at :func:`" + orig_module_name + "." + old_class_name + "." + decorated_name + "`."
        )

        decorated.__doc__ = replaced_descript + add_link_to_old_class

        return decorated

    return decorator


def diff_environment_yml(
    fn_env: str | dict[str, Any], fn_devenv: str | dict[str, Any], print_dep: str = "both", input_dict: bool = False
) -> None:
    """
    Compute the difference between environment.yml and dev-environment.yml for setup of continuous integration,
    while checking that all the dependencies listed in environment.yml are also in dev-environment.yml
    :param fn_env: Filename path to environment.yml
    :param fn_devenv: Filename path to dev-environment.yml
    :param print_dep: Whether to print conda differences "conda", pip differences "pip" or both.
    :param input_dict: Whether to consider the input as a dict (for testing purposes).
    """

    yaml = import_optional("yaml", package_name="pyyaml")

    if not input_dict:
        # Load the yml as dictionaries
        yaml_env = yaml.safe_load(open(fn_env))  # type: ignore
        yaml_devenv = yaml.safe_load(open(fn_devenv))  # type: ignore
    else:
        # We need a copy as we'll pop things out and don't want to affect input
        # dict.copy() is shallow and does not work with embedded list in dicts (as is the case here)
        yaml_env = copy.deepcopy(fn_env)
        yaml_devenv = copy.deepcopy(fn_devenv)

    # Extract the dependencies values
    conda_dep_env = yaml_env["dependencies"]
    conda_dep_devenv = yaml_devenv["dependencies"]

    # Check if there is any pip dependency, if yes pop it from the end of the list
    if isinstance(conda_dep_devenv[-1], dict):
        pip_dep_devenv = conda_dep_devenv.pop()["pip"]

        # Remove the package's self install for devs via pip, if it exists
        if "-e ./" in pip_dep_devenv:
            pip_dep_devenv.remove("-e ./")

        # Check if there is a pip dependency in the normal env as well, if yes pop it also
        if isinstance(conda_dep_env[-1], dict):
            pip_dep_env = conda_dep_env.pop()["pip"]

            # The diff below computes the dependencies that are in env but not in dev-env
            # It should be empty, otherwise we raise an error
            diff_pip_check = list(set(pip_dep_env) - set(pip_dep_devenv))
            if len(diff_pip_check) != 0:
                raise ValueError(
                    "The following pip dependencies are listed in env but not dev-env: " + ",".join(diff_pip_check)
                )

            # The diff below computes the dependencies that are in dev-env but not in env, to add during CI
            diff_pip_dep = list(set(pip_dep_devenv) - set(pip_dep_env))

        # If there is no pip dependency in env, all the ones of dev-env need to be added during CI
        else:
            diff_pip_dep = pip_dep_devenv

    # If there is no pip dependency, we ignore this step
    else:
        diff_pip_dep = []

    # If the diff is empty for pip, return a string "None" to read easily in bash
    if len(diff_pip_dep) == 0:
        diff_pip_dep = ["None"]

    # We do the same for the conda dependency, first a sanity check that everything that is in env is also in dev-ev
    diff_conda_check = list(set(conda_dep_env) - set(conda_dep_devenv))
    if len(diff_conda_check) != 0:
        raise ValueError("The following dependencies are listed in env but not dev-env: " + ",".join(diff_conda_check))

    # Then the difference to add during CI
    diff_conda_dep = list(set(conda_dep_devenv) - set(conda_dep_env))

    # Join the lists
    joined_list_conda_dep = " ".join(diff_conda_dep)
    joined_list_pip_dep = " ".join(diff_pip_dep)

    # Print to be captured in bash
    if print_dep == "both":
        print(joined_list_conda_dep)
        print(joined_list_pip_dep)
    elif print_dep == "conda":
        print(joined_list_conda_dep)
    elif print_dep == "pip":
        print(joined_list_pip_dep)
    else:
        raise ValueError('The argument "print_dep" can only be "conda", "pip" or "both".')



"""
Reproject a vector
==================

This example demonstrates the reprojection of a vector using :func:`geoutils.Vector.reproject`.
"""

# %%
# We open a raster and vector.

# sphinx_gallery_thumbnail_number = 3
import geoutils as gu

filename_rast = gu.examples.get_path("everest_landsat_b4_cropped")
filename_vect = gu.examples.get_path("everest_rgi_outlines")
rast = gu.Raster(filename_rast)
vect = gu.Vector(filename_vect)

# %%
# The two objects are in different projections.
rast.info()
vect.info()

# %%
# Let's plot the two in their original projection.
rast.plot(cmap="Greys_r")
vect.plot(ax="new", fc="none", ec="tab:purple", lw=3)

# %%
# **First option:** using the raster as a reference to match, we reproject the vector. We simply have to pass the :class:`~geoutils.Raster` as an argument
# to :func:`~geoutils.Vector.reproject`. See :ref:`core-match-ref` for more details.

vect_reproj = vect.reproject(rast)

# %%
# We can plot the vector in its new projection.

vect_reproj.plot(ax="new", fc="none", ec="tab:purple", lw=3)

# %%
# **Second option:** we can pass the georeferencing argument ``dst_crs`` to :func:`~geoutils.Vector.reproject` (an EPSG code can be passed directly as
# :class:`int`).

# Reproject in UTM zone 45N.
vect_reproj = vect.reproject(crs=32645)
vect_reproj



"""Test functions specific to the Vector class."""

from __future__ import annotations

import inspect
import os.path
import pathlib
import tempfile
import warnings

import geopandas as gpd
import geopandas.base
import matplotlib.pyplot as plt
import pyproj
import pytest
from geopandas.testing import assert_geodataframe_equal, assert_geoseries_equal
from pandas.testing import assert_series_equal
from shapely.geometry.base import BaseGeometry
from shapely.geometry.linestring import LineString
from shapely.geometry.polygon import Polygon

import geoutils as gu

DO_PLOT = False


class TestVector:
    landsat_b4_crop_path = gu.examples.get_path_test("everest_landsat_b4_cropped")
    everest_outlines_path = gu.examples.get_path_test("everest_rgi_outlines")
    aster_dem_path = gu.examples.get_path_test("exploradores_aster_dem")
    aster_outlines_path = gu.examples.get_path_test("exploradores_rgi_outlines")

    def test_init(self) -> None:
        """Test class initiation works as intended"""

        # First, with a string filename
        v0 = gu.Vector(self.aster_outlines_path)
        assert isinstance(v0, gu.Vector)

        # Second, with a pathlib path
        path = pathlib.Path(self.aster_outlines_path)
        v1 = gu.Vector(path)
        assert isinstance(v1, gu.Vector)

        # Third, with a geopandas dataframe
        v2 = gu.Vector(gpd.read_file(self.aster_outlines_path))
        assert isinstance(v2, gu.Vector)

        # Fourth, passing a Vector itself (points back to Vector passed)
        v3 = gu.Vector(v2)
        assert isinstance(v3, gu.Vector)

        # Check errors are raised when filename has wrong type
        with pytest.raises(TypeError, match="Filename argument should be a string, path or geodataframe."):
            gu.Vector(1)  # type: ignore

    def test_copy(self) -> None:

        vector = gu.Vector(self.aster_outlines_path)
        vector2 = vector.copy()

        # Assert they are not the same object
        assert vector2 is not vector

        # Modify vector2, and check vector is not affected
        vector2.ds = vector2.ds.query("RGIId == 'RGI60-17.08409'")
        assert vector2.ds.shape[0] < vector.ds.shape[0]

    def test_info(self) -> None:

        v = gu.Vector(self.aster_outlines_path)

        # Check default runs without error (prints to screen)
        output = v.info()
        assert output is None

        # Otherwise returns info
        output2 = v.info(verbose=False)
        assert isinstance(output2, str)
        list_prints = ["Filename", "Coordinate system", "Extent", "Number of features", "Attributes"]
        assert all(p in output2 for p in list_prints)

    def test_to_file(self) -> None:
        """Test the save wrapper for GeoDataFrame.to_file()."""

        vector = gu.Vector(self.aster_outlines_path)

        # Create a temporary file in a temporary directory
        temp_dir = tempfile.TemporaryDirectory()
        temp_file = os.path.join(temp_dir.name, "test.gpkg")

        # Save and check the file exists
        vector.to_file(temp_file)
        assert os.path.exists(temp_file)

        # Open and check the object is the same
        vector_save = gu.Vector(temp_file)
        vector_save.vector_equal(vector)

    def test_bounds(self) -> None:
        vector = gu.Vector(self.aster_outlines_path)
        bounds = vector.bounds

        assert bounds.left < bounds.right
        assert bounds.bottom < bounds.top

        assert bounds.left == vector.ds.total_bounds[0]
        assert bounds.bottom == vector.ds.total_bounds[1]
        assert bounds.right == vector.ds.total_bounds[2]
        assert bounds.top == vector.ds.total_bounds[3]

    def test_footprint(self) -> None:

        vector = gu.Vector(self.aster_outlines_path)
        footprint = vector.footprint

        assert isinstance(footprint, gu.Vector)
        assert footprint.vector_equal(vector.get_footprint_projected(vector.crs))

    def test_plot(self) -> None:
        """Test the vector plot."""

        # Load vector example
        vector = gu.Vector(self.aster_outlines_path)

        # Test default plot
        vector.plot()

        if DO_PLOT:
            plt.show()
        else:
            plt.close()
        assert True

        # Test with new figure
        plt.figure()
        vector.plot()
        if DO_PLOT:
            plt.show()
        else:
            plt.close()
        assert True

        # Test with provided ax
        ax = plt.subplot(111)
        vector.plot(ax=ax)
        if DO_PLOT:
            plt.show()
        else:
            plt.close()
        assert True

        # Test save fig
        temp_dir = tempfile.TemporaryDirectory()
        temp_file = os.path.join(temp_dir.name, "test.png")
        vector.plot(savefig_fname=temp_file)
        if DO_PLOT:
            plt.show()
        else:
            plt.close()
        assert os.path.isfile(temp_file)


class NeedToImplementWarning(FutureWarning):
    """Warning to remember to implement new GeoPandas methods"""


class TestGeoPandasMethods:
    # Use two synthetic vectors
    poly = Polygon([(10, 10), (11, 10), (11, 11), (10, 11)])
    gdf1 = gpd.GeoDataFrame({"geometry": [poly]}, crs="EPSG:4326")
    synthvec1 = gu.Vector(gdf1)

    # Create a synthetic LineString geometry
    lines = LineString([(10, 10), (10.5, 10.5), (11, 11)])
    gdf2 = gpd.GeoDataFrame({"geometry": [lines]}, crs="EPSG:4326")
    synthvec2 = gu.Vector(gdf2)

    # Use a real-life vector, divided in 2 for speed of tests
    realvec = gu.Vector(gu.examples.get_path_test("everest_rgi_outlines"))
    realvec1 = realvec[0:2]
    realvec2 = realvec[2:4]

    # Properties and methods derived from Shapely or GeoPandas
    # List of properties and methods with non-geometric output that are implemented in GeoUtils
    main_properties = ["crs", "geometry", "total_bounds", "active_geometry_name"]
    nongeo_properties = [
        "area",
        "length",
        "interiors",
        "geom_type",
        "is_empty",
        "is_ring",
        "is_simple",
        "is_valid",
        "has_z",
        "is_ccw",
        "is_closed",
    ]
    nongeo_methods = [
        "contains",
        "geom_equals",
        "geom_equals_exact",
        "crosses",
        "disjoint",
        "intersects",
        "overlaps",
        "touches",
        "within",
        "covers",
        "covered_by",
        "distance",
        "is_valid_reason",
        "count_coordinates",
        "count_geometries",
        "count_interior_rings",
        "get_precision",
        "minimum_clearance",
        "minimum_bounding_radius",
        "contains_properly",
        "dwithin",
        "hausdorff_distance",
        "frechet_distance",
        "hilbert_distance",
        "relate_pattern",
        "relate",
        "project",
    ]

    # List of properties and methods with geometric output that are implemented in GeoUtils
    geo_properties = ["boundary", "centroid", "convex_hull", "envelope", "exterior"]
    geo_methods = [
        "representative_point",
        "normalize",
        "make_valid",
        "difference",
        "symmetric_difference",
        "union",
        "union_all",
        "intersection",
        "clip_by_rect",
        "buffer",
        "simplify",
        "affine_transform",
        "translate",
        "rotate",
        "scale",
        "skew",
        "concave_hull",
        "delaunay_triangles",
        "voronoi_polygons",
        "minimum_rotated_rectangle",
        "minimum_bounding_circle",
        "extract_unique_points",
        "offset_curve",
        "remove_repeated_points",
        "reverse",
        "segmentize",
        "transform",
        "force_2d",
        "force_3d",
        "line_merge",
        "intersection_all",
        "snap",
        "build_area",
        "polygonize",
        "shortest_line",
        "dissolve",
        "explode",
        "sjoin",
        "sjoin_nearest",
        "overlay",
        "to_crs",
        "set_crs",
        "rename_geometry",
        "set_geometry",
        "set_precision",
        "get_geometry",
        "shared_paths",
        "clip",
        "interpolate",
    ]
    # List of class methods
    io_methods = [
        "from_file",
        "from_postgis",
        "from_dict",
        "from_features",
        "from_arrow",
        "to_feather",
        "to_parquet",
        "to_file",
        "to_postgis",
        "to_arrow",
        "to_geo_dict",
        "to_json",
        "to_wkb",
        "to_wkt",
        "to_csv",
    ]

    # List of other properties and methods
    other = ["has_sindex", "sindex", "estimate_utm_crs", "cx", "iterfeatures", "get_coordinates"]
    all_declared = (
        main_properties + nongeo_methods + nongeo_properties + geo_methods + geo_properties + other + io_methods
    )

    # Methods that are implemented but difficult to test, so we only check consistency of signature (name/args),
    # but we skip the test with placeholder data
    exceptions_skip_test = ["interpolate", "project", "shared_paths"]

    # Exceptions for GeoPandasBase functions not implemented (or deprecrated) in GeoSeries/GeoDataFrame
    exceptions_unimplemented = [
        "plot",  # Own implementation in Vector
        "explore",  # Not implemented in Vector
        "cascaded_union",  # Deprecated
        "unary_union",  # Deprecated
        "bounds",  # Own implementation in Vector
        "equals",
        "type",
        "convert_dtypes",
        "merge",
        "apply",
        "astype",
        "sample_points",
        "copy",
    ]
    # Exceptions for IO/conversion that can be done directly from .ds
    all_exceptions = exceptions_unimplemented

    # Get all GeoPandasBase public methods with some exceptions
    geobase_methods = gpd.base.GeoPandasBase.__dict__.copy()

    # Get all GeoDataFrame public methods with some exceptions
    gdf_methods = gpd.GeoDataFrame.__dict__.copy()

    def test_overridden_funcs_exist(self) -> None:
        """Check that all methods listed above exist in Vector."""

        # Check that all methods declared in the class above exist in Vector
        vector_methods = gu.Vector.__dict__

        list_missing = [method for method in self.all_declared if method not in vector_methods.keys()]

        assert len(list_missing) == 0, print(f"Test method listed that is not in GeoUtils: {list_missing}")

    def test_geopandas_coverage(self) -> None:
        """Check that all existing methods of GeoPandas are overridden, with a couple exceptions."""

        # Merge the two
        all_methods = self.geobase_methods.copy()
        all_methods.update(self.gdf_methods)

        # Remove exceptions we don't want to reuse from GeoPandas (mirrored in Vector)
        name_all_methods = list(all_methods.keys())
        public_methods = [method for method in name_all_methods if method[0] != "_"]

        covered_methods = [method for method in public_methods if method not in self.all_exceptions]

        # Check that all methods declared in the class above are covered in Vector
        list_missing = [method for method in covered_methods if method not in self.all_declared]

        if len(list_missing) != 0:
            warnings.warn(
                f"New GeoPandas methods are not implemented in GeoUtils: {list_missing}", NeedToImplementWarning
            )

    @pytest.mark.parametrize("method", nongeo_methods + geo_methods)  # type: ignore
    def test_overridden_funcs_args(self, method: str) -> None:
        """Check that all methods overridden have the same arguments as in GeoPandas."""

        # Skip methods with added in-place argument
        if method in ["translate", "set_precision"]:
            return
        # Get GeoPandas class where the methods live
        elif method in self.geobase_methods.keys():
            upstream_class = gpd.base.GeoPandasBase
        elif method in self.gdf_methods.keys():
            upstream_class = gpd.GeoDataFrame
        else:
            raise ValueError("Method did not belong to GeoDataFrame or GeoPandasBase class.")

        # Get a full argument inspection object for each class
        argspec_upstream = inspect.getfullargspec(getattr(upstream_class, method))
        argspec_geoutils = inspect.getfullargspec(getattr(gu.Vector, method))

        # Check that all positional arguments are the same
        if argspec_upstream.args != argspec_geoutils.args:
            warnings.warn("Argument of GeoPandas method not consistent in GeoUtils.", NeedToImplementWarning)

        # Check that the *args and **kwargs argument are declared consistently
        if argspec_upstream.varargs != argspec_geoutils.varargs:
            warnings.warn("Argument of GeoPandas method not consistent in GeoUtils.", NeedToImplementWarning)

        if argspec_upstream.varkw != argspec_geoutils.varkw:
            warnings.warn("Argument of GeoPandas method not consistent in GeoUtils.", NeedToImplementWarning)

        # Check that default argument values are the same
        if argspec_upstream.defaults != argspec_geoutils.defaults:
            warnings.warn("Default argument of GeoPandas method not consistent in GeoUtils.", NeedToImplementWarning)

    @pytest.mark.parametrize("vector", [synthvec1, synthvec2, realvec1, realvec2])  # type: ignore
    @pytest.mark.parametrize("method", nongeo_properties)  # type: ignore
    def test_nongeo_properties(self, vector: gu.Vector, method: str) -> None:
        """Check non-geometric properties are consistent with GeoPandas."""

        # Remove warnings about operations in a non-projected system, and future changes
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=FutureWarning)

        # Get method for each class
        output_geoutils = getattr(vector, method)
        output_geopandas = getattr(vector.ds, method)

        # Assert equality
        assert_series_equal(output_geoutils, output_geopandas)

    @pytest.mark.parametrize("vector1", [synthvec1, realvec1])  # type: ignore
    @pytest.mark.parametrize("vector2", [synthvec2, realvec2])  # type: ignore
    @pytest.mark.parametrize("method", nongeo_methods)  # type: ignore
    def test_nongeo_methods(self, vector1: gu.Vector, vector2: gu.Vector, method: str) -> None:
        """
        Check non-geometric methods are consistent with GeoPandas.
        Many of these methods require two inputs ("other", "df", or "right" argument), except a few.
        """

        # Remove warnings about operations in a non-projected system, and future changes
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=FutureWarning)

        # Get method for each class

        # Methods with no input
        if method in self.exceptions_skip_test:
            return
        elif method in [
            "is_valid_reason",
            "count_coordinates",
            "count_geometries",
            "count_interior_rings",
            "get_precision",
            "minimum_clearance",
            "minimum_bounding_radius",
            "hilbert_distance",
        ]:
            output_geoutils = getattr(vector1, method)()
            output_geopandas = getattr(vector1.ds, method)()
        elif method == "geom_equals_exact":
            output_geoutils = getattr(vector1, method)(vector2, tolerance=0.1)
            output_geopandas = getattr(vector1.ds, method)(vector2.ds, tolerance=0.1)
        elif method == "dwithin":
            output_geoutils = getattr(vector1, method)(vector2, distance=0.1)
            output_geopandas = getattr(vector1.ds, method)(vector2.ds, distance=0.1)
        elif method == "relate_pattern":
            output_geoutils = getattr(vector1, method)(vector2, pattern="*T*******")
            output_geopandas = getattr(vector1.ds, method)(vector2.ds, pattern="*T*******")
        else:
            output_geoutils = getattr(vector1, method)(vector2)
            output_geopandas = getattr(vector1.ds, method)(vector2.ds)

        # Assert equality
        assert_series_equal(output_geoutils, output_geopandas)

    @pytest.mark.parametrize("vector", [synthvec1, synthvec2, realvec1, realvec2])  # type: ignore
    @pytest.mark.parametrize("method", geo_properties)  # type: ignore
    def test_geo_properties(self, vector: gu.Vector, method: str) -> None:
        """Check geometric properties are consistent with GeoPandas."""

        # Remove warnings about operations in a non-projected system, and future changes
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=FutureWarning)

        # Get method for each class
        output_geoutils = getattr(vector, method)
        output_geopandas = getattr(vector.ds, method)

        # Assert output types
        assert isinstance(output_geoutils, gu.Vector)
        assert isinstance(output_geopandas, (gpd.GeoSeries, gpd.GeoDataFrame, BaseGeometry))

        # Separate cases depending on GeoPandas' output
        if isinstance(output_geopandas, gpd.GeoSeries):
            # Assert geoseries equality
            assert_geoseries_equal(output_geoutils.ds.geometry, output_geopandas)
        elif isinstance(output_geopandas, BaseGeometry):
            assert_geodataframe_equal(
                output_geoutils.ds, gpd.GeoDataFrame({"geometry": [output_geopandas]}, crs=vector.crs)
            )
        else:
            assert_geodataframe_equal(output_geoutils.ds, output_geopandas)

    specific_method_args = {
        "buffer": {"distance": 1},
        "clip_by_rect": {"xmin": 10.5, "ymin": 10.5, "xmax": 11, "ymax": 11},
        "affine_transform": {"matrix": [1, 1, 1, 1, 1, 1]},
        "translate": {"xoff": 1, "yoff": 1, "zoff": 0},
        "rotate": {"angle": 90},
        "scale": {"xfact": 1.1, "yfact": 1.1, "zfact": 1.1, "origin": "center"},
        "skew": {"xs": 1.1, "ys": 1.1},
        "interpolate": {"distance": 1},
        "simplify": {"tolerance": 0.1},
        "to_crs": {"crs": pyproj.CRS.from_epsg(32610)},
        "set_crs": {"crs": pyproj.CRS.from_epsg(32610), "allow_override": True},
        "rename_geometry": {"col": "lol"},
        "set_geometry": {"col": synthvec1.geometry},
        "set_precision": {"grid_size": 1},
        "clip": {"mask": poly},
        "concave_hull": {"ratio": 0.5},
        "delaunay_triangles": {"tolerance": 0.1},
        "voronoi_polygons": {"tolerance": 0.0},
        "offset_curve": {"distance": 0.1},
        "remove_repeated_points": {"tolerance": 0},
        "segmentize": {"max_segment_length": 0.1},
        "transform": {"transformation": lambda x: x + 1},
        "get_geometry": {"index": 1},
    }

    @pytest.mark.parametrize("vector1", [synthvec1, realvec1])  # type: ignore
    @pytest.mark.parametrize("vector2", [synthvec2, realvec2])  # type: ignore
    @pytest.mark.parametrize("method", geo_methods)  # type: ignore
    def test_geo_methods(self, vector1: gu.Vector, vector2: gu.Vector, method: str) -> None:
        """Check geometric methods are consistent with GeoPandas."""

        # Remove warnings about operations in a non-projected system, and future changes
        warnings.simplefilter("ignore", category=UserWarning)
        warnings.simplefilter("ignore", category=FutureWarning)

        # Methods that require two inputs
        if method in self.exceptions_skip_test:
            return
        elif method in [
            "difference",
            "symmetric_difference",
            "union",
            "intersection",
            "sjoin",
            "sjoin_nearest",
            "overlay",
            "shortest_line",
        ]:
            output_geoutils = getattr(vector1, method)(vector2)
            output_geopandas = getattr(vector1.ds, method)(vector2.ds)
        # Methods that require zero input
        elif method in [
            "representative_point",
            "normalize",
            "make_valid",
            "dissolve",
            "explode",
            "minimum_rotated_rectangle",
            "minimum_bounding_circle",
            "extract_unique_points",
            "reverse",
            "force_2d",
            "force_3d",
            "intersection_all",
            "union_all",
            "build_area",
            "polygonize",
            "line_merge",
        ]:
            output_geoutils = getattr(vector1, method)()
            output_geopandas = getattr(vector1.ds, method)()
        elif method in ["snap"]:
            output_geoutils = getattr(vector1, method)(vector2, tolerance=0.1)
            output_geopandas = getattr(vector1.ds, method)(vector2.ds, tolerance=0.1)
        elif method in self.specific_method_args.keys():
            output_geoutils = getattr(vector1, method)(**self.specific_method_args[method])
            output_geopandas = getattr(vector1.ds, method)(**self.specific_method_args[method])
        else:
            raise ValueError(f"The method '{method}' is not covered by this test.")

        # Assert output types
        assert isinstance(output_geoutils, gu.Vector)
        assert isinstance(output_geopandas, (BaseGeometry, gpd.GeoSeries, gpd.GeoDataFrame))

        # Separate cases depending on GeoPandas' output, and nature of the function
        # Simplify is a special case that can make geometries invalid, so adjust test
        if method == "simplify":
            # TODO: Unskip this random test failure (one index not matching) when this is fixed in GeoPandas/Shapely
            pass
            # assert_geoseries_equal(
            #     output_geopandas.make_valid(), output_geoutils.ds.geometry.make_valid(), check_less_precise=True
            # )
        elif isinstance(output_geopandas, BaseGeometry):
            output_geopandas.equals(output_geoutils.ds.geometry)
        # For geoseries output, check equality of it
        elif isinstance(output_geopandas, gpd.GeoSeries):
            assert_geoseries_equal(output_geoutils.ds.geometry, output_geopandas)
        # For geodataframe output, check equality
        else:
            assert_geodataframe_equal(output_geoutils.ds, output_geopandas)



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functionalities for gridding points (point cloud to raster)."""

import warnings
from typing import Literal

import affine
import geopandas as gpd
import numpy as np
import rasterio as rio
from scipy.interpolate import griddata

from geoutils._typing import NDArrayNum


def _grid_pointcloud(
    pc: gpd.GeoDataFrame,
    grid_coords: tuple[NDArrayNum, NDArrayNum] = None,
    data_column_name: str | None = None,
    resampling: Literal["nearest", "linear", "cubic"] = "linear",
    dist_nodata_pixel: float = 1.0,
) -> tuple[NDArrayNum, affine.Affine]:
    """
    Grid point cloud (possibly irregular coordinates) to raster (regular grid) using delaunay triangles interpolation.

    Based on scipy.interpolate.griddata combined to a nearest point search to replace values of grid cells further than
    a certain distance (in number of pixels) by nodata values (as griddata interpolates all values in convex hull, no
    matter the distance).

    :param pc: Point cloud.
    :param grid_coords: Regular raster grid coordinates in X and Y (i.e. equally spaced, independently for each axis).
    :param data_column_name: Name of data column for point cloud (if 2D point geometries are used).
    :param resampling: Resampling method within delauney triangles (defaults to linear).
    :param dist_nodata_pixel: Distance from the point cloud after which grid cells are filled by nodata values,
        expressed in number of pixels.
    """

    # Input checks
    if (
        not isinstance(grid_coords, tuple)
        or not (isinstance(grid_coords[0], np.ndarray) and grid_coords[0].ndim == 1)
        or not (isinstance(grid_coords[1], np.ndarray) and grid_coords[1].ndim == 1)
    ):
        raise TypeError("Input grid coordinates must be 1D arrays.")

    diff_x = np.diff(grid_coords[0])
    diff_y = np.diff(grid_coords[1])

    if not all(diff_x == diff_x[0]) and all(diff_y == diff_y[0]):
        raise ValueError("Grid coordinates must be regular (equally spaced, independently along X and Y).")

    # 1/ Interpolate irregular point cloud on a regular grid

    # Get meshgrid coordinates
    xx, yy = np.meshgrid(grid_coords[0], grid_coords[1])

    # Use griddata on all points
    aligned_dem = griddata(
        points=(pc.geometry.x.values, pc.geometry.y.values),
        values=pc[data_column_name].values if data_column_name is not None else pc.geometry.z.values,
        xi=(xx, yy),
        method=resampling,
        rescale=True,  # Rescale inputs to unit cube to avoid precision issues
    )

    # 2/ Identify which grid points are more than X pixels away from the point cloud, and convert to NaNs
    # (otherwise all grid points in the convex hull of the irregular triangulation are filled, no matter the distance)

    # Get the nearest point for each grid point
    grid_pc = gpd.GeoDataFrame(
        data={"placeholder": np.ones(len(xx.ravel()))},
        geometry=gpd.points_from_xy(x=xx.ravel(), y=yy.ravel()),
        crs=pc.crs,
    )
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, message="Geometry is in a geographic CRS.*")
        near = gpd.sjoin_nearest(grid_pc, pc)
        # In case there are several points at the same distance, it doesn't matter which one is used to compute the
        # distance, so we keep the first index of closest point
        index_right = near.groupby(by=near.index)["index_right"].min()

    # Compute distance between points as a function of the pixel sizes in X and Y
    res_x = np.abs(grid_coords[0][1] - grid_coords[0][0])
    res_y = np.abs(grid_coords[1][1] - grid_coords[1][0])
    dist = np.sqrt(
        ((pc.geometry.x.values[index_right] - grid_pc.geometry.x.values) / res_x) ** 2
        + ((pc.geometry.y.values[index_right] - grid_pc.geometry.y.values) / res_y) ** 2
    )

    # Replace all points further away than the distance of nodata by NaNs
    aligned_dem[dist.reshape(aligned_dem.shape) > dist_nodata_pixel] = np.nan

    # Flip Y axis of grid
    aligned_dem = np.flip(aligned_dem, axis=0)

    # 3/ Derive output transform from input grid
    transform_from_coords = rio.transform.from_origin(min(grid_coords[0]), max(grid_coords[1]), res_x, res_y)

    return aligned_dem, transform_from_coords



"""
Test the example files used for testing and documentation
"""

import hashlib
import warnings

import pytest

import geoutils as gu
from geoutils import examples


@pytest.mark.parametrize(
    "example", ["everest_landsat_b4", "everest_landsat_b4_cropped", "everest_landsat_rgb", "exploradores_aster_dem"]
)  # type: ignore
def test_read_paths_raster(example: str) -> None:
    assert isinstance(gu.Raster(examples.get_path(example)), gu.Raster)
    assert isinstance(gu.Raster(examples.get_path_test(example)), gu.Raster)


@pytest.mark.parametrize("example", ["everest_rgi_outlines", "exploradores_rgi_outlines"])  # type: ignore
def test_read_paths_vector(example: str) -> None:
    warnings.simplefilter("error")
    assert isinstance(gu.Vector(examples.get_path(example)), gu.Vector)
    assert isinstance(gu.Vector(examples.get_path_test(example)), gu.Vector)


# Original sha256 obtained with `sha256sum filename`
original_sha256_examples = {
    "everest_landsat_b4": "271fa34e248f016f87109c8e81960caaa737558fbae110ec9e0d9e2d30d80c26",
    "everest_landsat_b4_cropped": "0e63d8e9c4770534a1ec267c91e80cd9266732184a114f0bd1aadb5a613215e6",
    "everest_landsat_rgb": "7d0505a8610fd7784cb71c03e5b242715cd1574e978c2c86553d60fd82372c30",
    "everest_rgi_outlines": "d1a5bcd4bd4731a24c2398c016a6f5a8064160fedd5bab10609adacda9ba41ef",
    "exploradores_aster_dem": "dcb0d708d042553cdd2bb4fd82c55b5674a5e0bd6ea46f1a021b396b7d300033",
    "exploradores_rgi_outlines": "19c2dac089ce57373355213fdf2fd72f601bf97f21b04c4920edb1e4384ae2b2",
    "coromandel_lidar": "2f1fff1bb84860a8438e14d39e14bf974236dc6345e64649a131507d0ed844f3",
}


@pytest.mark.parametrize("example", examples.available)  # type: ignore
def test_data_integrity__examples(example: str) -> None:
    """
    Test that input data is not corrupted by checking sha265 sum
    """
    # Read file as bytes
    fbytes = open(examples.get_path(example), "rb").read()

    # Get sha256
    file_sha256 = hashlib.sha256(fbytes).hexdigest()

    assert file_sha256 == original_sha256_examples[example]


original_sha256_test = {
    "everest_landsat_b4": "5aa1a0a1c17efd211e42218ab5e2f3e0e404b96ba5055ac5eebef756ad5c65bc",
    "everest_landsat_b4_cropped": "4244767c31c51f7c7b5fb8eb48df7d6394aa707deb9fe699d5672ff9d2507aef",
    "everest_landsat_rgb": "b77109f8027418cdd36ccab34cc3996bbbf2756b116ecf0fed8e4163cd7aa2f9",
    "everest_rgi_outlines": "3642e2fa5da1d9cad2378e0941985ae47077dba7839e30fdd413f8e868ab2ade",
    "exploradores_aster_dem": "c98f24cb131810dd8b2f4773a8df0821cf31edec79890967a67b8a6fdb89314d",
    "exploradores_rgi_outlines": "2f0281b00a49ad2f0874fb4ee54df1e0d11ad073f826d9ca713430588c15fa15",
    "coromandel_lidar": "95af5de14205c712e7674723d00119f4fa6239a65fb2aa3f7035254ace3194ae",
}


@pytest.mark.parametrize("example_test", examples.available_test)  # type: ignore
def test_data_integrity__tests(example_test: str) -> None:
    """
    Test that input data is not corrupted by checking sha265 sum
    """
    # Read file as bytes
    fbytes = open(examples.get_path_test(example_test), "rb").read()

    # Get sha256
    file_sha256 = hashlib.sha256(fbytes).hexdigest()

    assert file_sha256 == original_sha256_test[example_test]



# Copyright (c) 2025 GeoUtils developers
#
# This file is part of the GeoUtils project:
# https://github.com/glaciohack/geoutils
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for additional statistical estimators."""

from typing import Any

import numpy as np
from scipy.stats.mstats import mquantiles

from geoutils._typing import NDArrayNum


def nmad(data: NDArrayNum, nfact: float = 1.4826) -> np.floating[Any]:
    """
    Calculate the normalized median absolute deviation (NMAD) of an array.
    Default scaling factor is 1.4826 to scale the median absolute deviation (MAD) to the dispersion of a normal
    distribution (see https://en.wikipedia.org/wiki/Median_absolute_deviation#Relation_to_standard_deviation, and
    e.g. Höhle and Höhle (2009), http://dx.doi.org/10.1016/j.isprsjprs.2009.02.003)

    :param data: Input array or raster
    :param nfact: Normalization factor for the data

    :returns nmad: (normalized) median absolute deviation of data.
    """
    if isinstance(data, np.ma.masked_array):
        return nfact * np.ma.median(np.abs(data - np.ma.median(data)))
    else:
        return nfact * np.nanmedian(np.abs(data - np.nanmedian(data)))


def linear_error(data: NDArrayNum, interval: float = 90) -> np.floating[Any]:
    """
    Compute the linear error (LE) for a given dataset, representing the range of differences between the upper and
    lower percentiles of the data. By default, this calculates the 90% confidence interval (LE90).

    :param data: A numpy array or masked array of data, typically representing the differences (errors) in elevation or
    another quantity.
    :param interval: The confidence interval to compute, specified as a percentage. For example, an interval of 90 will
    compute the range between the 5th and 95th percentiles (LE90). This value must be between 0 and 100.

    return: The computed linear error, which is the difference between the upper and lower percentiles.

    raises: ValueError if the `interval` is not between 0 and 100.
    """
    # Validate the interval
    if not (0 < interval <= 100):
        raise ValueError("Interval must be between 0 and 100")

    max = 50 + interval / 2
    min = 50 - interval / 2
    if isinstance(data, np.ma.masked_array):
        return (
            mquantiles(data, prob=max / 100, alphap=1, betap=1) - mquantiles(data, prob=min / 100, alphap=1, betap=1)
        )[0]
    else:
        return np.nanpercentile(data, max) - np.nanpercentile(data, min)


def sum_square(data: NDArrayNum) -> np.floating[Any]:
    """
    Calculate the sum of the square of a data array.

    :param data: A numpy array or masked array of data, typically representing the differences (errors) in elevation or
    another quantity.
    :return: sum square
    """
    if np.ma.isMaskedArray(data):
        return np.ma.sum(np.square(data))
    else:
        return np.nansum(np.square(data))


def rmse(data: NDArrayNum) -> np.floating[Any]:
    """
    Calculate the RMSE of a data array.

    :param data: A numpy array or masked array of data, typically representing the differences (errors) in elevation or
    another quantity.
    :return: rmse
    """
    if np.ma.isMaskedArray(data):
        return np.sqrt(np.ma.mean(np.square(data)))
    else:
        return np.sqrt(np.nanmean(np.square(data)))
